# Copyright (c) OpenMMLab. All rights reserved.

from mmcv.cnn import MODELS as MMCV_MODELS
from mmcv.utils import Registry

from .hybrik import HybrIK_trainer
from .mesh_estimator import ImageBodyModelEstimator, VideoBodyModelEstimator


def build_from_cfg(cfg, registry, default_args=None):
    if cfg is None:
        return None
    return MMCV_MODELS.build_func(cfg, registry, default_args)


ARCHITECTURES = Registry(
    'architectures', parent=MMCV_MODELS, build_func=build_from_cfg)

ARCHITECTURES.register_module(name='HybrIK_trainer', module=HybrIK_trainer)
ARCHITECTURES.register_module(
    name='ImageBodyModelEstimator', module=ImageBodyModelEstimator)
ARCHITECTURES.register_module(
    name='VideoBodyModelEstimator', module=VideoBodyModelEstimator)


def build_architecture(cfg):
    """Build framework."""
    return ARCHITECTURES.build(cfg)



from typing import Optional

import torch
import torch.nn as nn
from pytorch3d.ops import interpolate_face_attributes
from pytorch3d.renderer import BlendParams, hard_rgb_blend
from pytorch3d.renderer.mesh.shader import SoftSilhouetteShader
from pytorch3d.structures.utils import padded_to_packed


class SilhouetteShader(SoftSilhouetteShader):
    """Avoid unexpected keyword argument error."""

    def __init__(self,
                 blend_params: Optional[BlendParams] = None,
                 **kwargs) -> None:
        super().__init__(blend_params)


class NoLightShader(nn.Module):
    """No light shader."""

    def __init__(self,
                 blend_params: Optional[BlendParams] = None,
                 **kwargs) -> None:
        """Initlialize without blend_params."""
        super().__init__()
        self.blend_params = blend_params if blend_params is not None\
            else BlendParams()

    def forward(self, fragments, meshes, **kwargs) -> torch.Tensor:
        """Sample without light."""
        texels = meshes.sample_textures(fragments)
        blend_params = kwargs.get('blend_params', self.blend_params)
        images = hard_rgb_blend(texels, fragments, blend_params)
        return images


class DepthShader(nn.Module):
    """No light shader."""

    def __init__(self,
                 blend_params: Optional[BlendParams] = None,
                 **kwargs) -> None:
        """Initlialize without blend_params."""
        super().__init__()
        self.blend_params = blend_params if blend_params is not None\
            else BlendParams()

    def forward(self, fragments, meshes, cameras, **kwargs) -> torch.Tensor:
        """Sample without light."""
        verts_depth = cameras.compute_depth_of_points(meshes.verts_padded())
        faces = meshes.faces_packed()  # (F, 3)
        verts_depth = padded_to_packed(verts_depth)
        faces_depth = verts_depth[faces]
        depth_map = interpolate_face_attributes(
            pix_to_face=fragments.pix_to_face,
            barycentric_coords=fragments.bary_coords,
            face_attributes=faces_depth)
        return depth_map[..., 0, :]


class NormalShader(nn.Module):
    """No light shader."""

    def __init__(self,
                 blend_params: Optional[BlendParams] = None,
                 **kwargs) -> None:
        """Initlialize without blend_params."""
        super().__init__()
        self.blend_params = blend_params if blend_params is not None\
            else BlendParams()

    def forward(self, fragments, meshes, cameras, **kwargs) -> torch.Tensor:
        """Sample without light."""
        verts_normal = cameras.compute_normal_of_meshes(meshes)
        faces = meshes.faces_packed()  # (F, 3)
        verts_normal = padded_to_packed(verts_normal)
        faces_normal = verts_normal[faces]
        normal_map = interpolate_face_attributes(
            pix_to_face=fragments.pix_to_face,
            barycentric_coords=fragments.bary_coords,
            face_attributes=faces_normal)
        return normal_map[..., 0, :]


class SegmentationShader(nn.Module):
    """No light shader."""

    def __init__(self,
                 blend_params: Optional[BlendParams] = None,
                 **kwargs) -> None:
        """Initlialize without blend_params."""
        super().__init__()
        self.blend_params = blend_params if blend_params is not None\
            else BlendParams()

    def forward(self, fragments, meshes, **kwargs) -> torch.Tensor:
        """Sample without light."""
        verts_class = meshes.textures.verts_features_padded()
        faces = meshes.faces_packed()  # (F, 3)
        verts_class = padded_to_packed(verts_class)
        faces_class = verts_class[faces]
        segmentation_map = interpolate_face_attributes(
            pix_to_face=fragments.pix_to_face,
            barycentric_coords=fragments.bary_coords,
            face_attributes=faces_class).long()
        return segmentation_map[..., :, 0]



# ORIGINAL_NAMES = [
#     'head_top',             # 00, extrapolate 02-01
#     'head_center',          # 01
#     'neck',                 # 02
#     'right_clavicle',       # 03
#     'right_shoulder',       # 04
#     'right_elbow',          # 05
#     'right_wrist',          # 06
#     'left_clavicle',        # 07
#     'left_shoulder',        # 08
#     'left_elbow',           # 09
#     'left_wrist',           # 10
#     'spine0',               # 11
#     'spine1',               # 12
#     'spine2',               # 13
#     'spine3',               # 14
#     'spine4',               # 15
#     'right_hip',            # 16
#     'right_knee',           # 17
#     'right_ankle',          # 18
#     'left_hip',             # 19
#     'left_knee',            # 20
#     'left_ankle',           # 21
#     'SKEL_ROOT',            # 22
#     'FB_R_Brow_Out_000',    # 23
#     'SKEL_L_Toe0',          # 24
#     'MH_R_Elbow',           # 25
#     'SKEL_L_Finger01',      # 26
#     'SKEL_L_Finger02',      # 27
#     'SKEL_L_Finger31',      # 28
#     'SKEL_L_Finger32',      # 29
#     'SKEL_L_Finger41',      # 30
#     'SKEL_L_Finger42',      # 31
#     'SKEL_L_Finger11',      # 32
#     'SKEL_L_Finger12',      # 33
#     'SKEL_L_Finger21',      # 34
#     'SKEL_L_Finger22',      # 35
#     'RB_L_ArmRoll',         # 36
#     'IK_R_Hand',            # 37
#     'RB_R_ThighRoll',       # 38
#     'FB_R_Lip_Corner_000',  # 39
#     'SKEL_Pelvis',          # 40
#     'IK_Head',              # 41
#     'MH_R_Knee',            # 42
#     'FB_LowerLipRoot_000',  # 43
#     'FB_R_Lip_Top_000',     # 44
#     'FB_R_CheekBone_000',   # 45
#     'FB_UpperLipRoot_000',  # 46
#     'FB_L_Lip_Top_000',     # 47
#     'FB_LowerLip_000',      # 48
#     'SKEL_R_Toe0',          # 49
#     'FB_L_CheekBone_000',   # 50
#     'MH_L_Elbow',           # 51
#     'RB_L_ThighRoll',       # 52
#     'PH_R_Foot',            # 53
#     'FB_L_Eye_000',         # 54
#     'SKEL_L_Finger00',      # 55
#     'SKEL_L_Finger10',      # 56
#     'SKEL_L_Finger20',      # 57
#     'SKEL_L_Finger30',      # 58
#     'SKEL_L_Finger40',      # 59
#     'FB_R_Eye_000',         # 60
#     'PH_R_Hand',            # 61
#     'FB_L_Lip_Corner_000',  # 62
#     'IK_R_Foot',            # 63
#     'RB_Neck_1',            # 64
#     'IK_L_Hand',            # 65
#     'RB_R_ArmRoll',         # 66
#     'FB_Brow_Centre_000',   # 67
#     'FB_R_Lid_Upper_000',   # 68
#     'RB_R_ForeArmRoll',     # 69
#     'FB_L_Lid_Upper_000',   # 70
#     'MH_L_Knee',            # 71
#     'FB_Jaw_000',           # 72
#     'FB_L_Lip_Bot_000',     # 73
#     'FB_Tongue_000',        # 74
#     'FB_R_Lip_Bot_000',     # 75
#     'IK_Root',              # 76
#     'PH_L_Foot',            # 77
#     'FB_L_Brow_Out_000',    # 78
#     'SKEL_R_Finger00',      # 79
#     'SKEL_R_Finger10',      # 80
#     'SKEL_R_Finger20',      # 81
#     'SKEL_R_Finger30',      # 82
#     'SKEL_R_Finger40',      # 83
#     'PH_L_Hand',            # 84
#     'RB_L_ForeArmRoll',     # 85
#     'FB_UpperLip_000',      # 86
#     'SKEL_R_Finger01',      # 87
#     'SKEL_R_Finger02',      # 88
#     'SKEL_R_Finger31',      # 89
#     'SKEL_R_Finger32',      # 90
#     'SKEL_R_Finger41',      # 91
#     'SKEL_R_Finger42',      # 92
#     'SKEL_R_Finger11',      # 93
#     'SKEL_R_Finger12',      # 94
#     'SKEL_R_Finger21',      # 95
#     'SKEL_R_Finger22',      # 96
#     'FACIAL_facialRoot',    # 97
#     'IK_L_Foot',            # 98
#     'interpolated_nose'     # 99, mid-point of 45-50
# ]

GTA_KEYPOINTS = [
    'gta_head_top',  # 00
    'head',  # 01 - head_center
    'neck',  # 02 - neck
    'gta_right_clavicle',  # 03
    'right_shoulder',  # 04  - right_shoulder
    'right_elbow',  # 05  - right_elbow
    'right_wrist',  # 06  - right_wrist
    'gta_left_clavicle',  # 07
    'left_shoulder',  # 08  - left_shoulder
    'left_elbow',  # 09  - left_elbow
    'left_wrist',  # 10  - left_wrist
    'spine_2',  # 11  - spine0
    'gta_spine1',  # 12
    'spine_1',  # 13  - spine2
    'pelvis',  # 14  - pelvis
    'gta_spine4',  # 15
    'right_hip',  # 16  - right_hip
    'right_knee',  # 17  - right_knee
    'right_ankle',  # 18  - right_ankle
    'left_hip',  # 19  - left_hip
    'left_knee',  # 20  - left_knee
    'left_ankle',  # 21  - left_ankle
    'gta_SKEL_ROOT',  # 22
    'gta_FB_R_Brow_Out_000',  # 23
    'left_foot',  # 24  - SKEL_L_Toe0
    'gta_MH_R_Elbow',  # 25
    'left_thumb_2',  # 26  - SKEL_L_Finger01
    'left_thumb_3',  # 27  - SKEL_L_Finger02
    'left_ring_2',  # 28  - SKEL_L_Finger31
    'left_ring_3',  # 29  - SKEL_L_Finger32
    'left_pinky_2',  # 30  - SKEL_L_Finger41
    'left_pinky_3',  # 31  - SKEL_L_Finger42
    'left_index_2',  # 32  - SKEL_L_Finger11
    'left_index_3',  # 33  - SKEL_L_Finger12
    'left_middle_2',  # 34  - SKEL_L_Finger21
    'left_middle_3',  # 35  - SKEL_L_Finger22
    'gta_RB_L_ArmRoll',  # 36
    'gta_IK_R_Hand',  # 37
    'gta_RB_R_ThighRoll',  # 38
    'gta_FB_R_Lip_Corner_000',  # 39
    'gta_SKEL_Pelvis',  # 40
    'gta_IK_Head',  # 41
    'gta_MH_R_Knee',  # 42
    'gta_FB_LowerLipRoot_000',  # 43
    'gta_FB_R_Lip_Top_000',  # 44
    'gta_FB_R_CheekBone_000',  # 45
    'gta_FB_UpperLipRoot_000',  # 46
    'gta_FB_L_Lip_Top_000',  # 47
    'gta_FB_LowerLip_000',  # 48
    'right_foot',  # 49  - SKEL_R_Toe0
    'gta_FB_L_CheekBone_000',  # 50
    'gta_MH_L_Elbow',  # 51
    'gta_RB_L_ThighRoll',  # 52
    'gta_PH_R_Foot',  # 53
    'left_eye',  # 54  - FB_L_Eye_000
    'gta_SKEL_L_Finger00',  # 55
    'left_index_1',  # 56  - SKEL_L_Finger10
    'left_middle_1',  # 57  - SKEL_L_Finger20
    'left_ring_1',  # 58  - SKEL_L_Finger30
    'left_pinky_1',  # 59  - SKEL_L_Finger40
    'right_eye',  # 60  - FB_R_Eye_000
    'gta_PH_R_Hand',  # 61
    'gta_FB_L_Lip_Corner_000',  # 62
    'gta_IK_R_Foot',  # 63
    'gta_RB_Neck_1',  # 64
    'gta_IK_L_Hand',  # 65
    'gta_RB_R_ArmRoll',  # 66
    'gta_FB_Brow_Centre_000',  # 67
    'gta_FB_R_Lid_Upper_000',  # 68
    'gta_RB_R_ForeArmRoll',  # 69
    'gta_FB_L_Lid_Upper_000',  # 70
    'gta_MH_L_Knee',  # 71
    'gta_FB_Jaw_000',  # 72
    'gta_FB_L_Lip_Bot_000',  # 73
    'gta_FB_Tongue_000',  # 74
    'gta_FB_R_Lip_Bot_000',  # 75
    'gta_IK_Root',  # 76
    'gta_PH_L_Foot',  # 77
    'gta_FB_L_Brow_Out_000',  # 78
    'gta_SKEL_R_Finger00',  # 79
    'right_index_1',  # 80  - SKEL_R_Finger10
    'right_middle_1',  # 81  - SKEL_R_Finger20
    'right_ring_1',  # 82  - SKEL_R_Finger30
    'right_pinky_1',  # 83  - SKEL_R_Finger40
    'gta_PH_L_Hand',  # 84
    'gta_RB_L_ForeArmRoll',  # 85
    'gta_FB_UpperLip_000',  # 86
    'right_thumb_2',  # 87  - SKEL_R_Finger01
    'right_thumb_3',  # 88  - SKEL_R_Finger02
    'right_ring_2',  # 89  - SKEL_R_Finger31
    'right_ring_3',  # 90  - SKEL_R_Finger32
    'right_pinky_2',  # 91  - SKEL_R_Finger41
    'right_pinky_3',  # 92  - SKEL_R_Finger42
    'right_index_2',  # 93  - SKEL_R_Finger11
    'right_index_3',  # 94  - SKEL_R_Finger12
    'right_middle_2',  # 95  - SKEL_R_Finger21
    'right_middle_3',  # 96  - SKEL_R_Finger22
    'gta_FACIAL_facialRoot',  # 97
    'gta_IK_L_Foot',  # 98
    'nose'  # 99  - interpolated nose
]



from .builder import build_post_processing
from .smooth.gaus1d_filter import Gaus1dFilter
from .smooth.oneeuro_filter import OneEuroFilter
from .smooth.savgol_filter import SGFilter
from .smooth.smoothnet import SmoothNetFilter
from .speed_up.deciwatch import DeciWatchPostProcessing

__all__ = [
    'build_post_processing',
    'OneEuroFilter',
    'SGFilter',
    'Gaus1dFilter',
    'SmoothNetFilter',
    'DeciWatchPostProcessing',
]



# Copyright (c) OpenMMLab. All rights reserved.

from mmcv.utils import Registry

from .hrnet import PoseHighResolutionNet
from .resnet import ResNet, ResNetV1d

BACKBONES = Registry('backbones')

BACKBONES.register_module(name='ResNet', module=ResNet)
BACKBONES.register_module(name='ResNetV1d', module=ResNetV1d)
BACKBONES.register_module(
    name='PoseHighResolutionNet', module=PoseHighResolutionNet)


def build_backbone(cfg):
    """Build backbone."""
    if cfg is None:
        return None
    return BACKBONES.build(cfg)



import json
import os

from mmhuman3d.core.cameras.camera_parameters import CameraParameter
from mmhuman3d.utils.path_utils import check_path_suffix
from .renderer.vedo_render import VedoRenderer


def visualize_chessboard_kinects_rgb(chessboard_path: str,
                                     interactive: bool = True,
                                     show: bool = True):
    """Visualize all the RGB cameras in a chessboard file.

    Args:
        chessboard_path (str):
            Path to the chessboard file.
        interactive (bool, optional):
            Pause and interact with window (True) or
            continue execution (False).
            Defaults to True.
        show (bool, optional):
            Whether to show in a window.
            Defaults to True.
    """
    # Load camera parameter from a json file
    camera_para_json_dict = json.load(open(chessboard_path))
    camera_para_dict = {}
    for camera_id in camera_para_json_dict.keys():
        try:
            camera_id_int = int(camera_id)
            # if camera_id is an instance of int
            # and it can be divided by 2, it's an rgb camera
            if camera_id_int % 2 == 0:
                pass
            else:
                continue
        except ValueError:
            continue
        temp_camera_parameter = CameraParameter(name=camera_id)
        temp_camera_parameter.load_from_chessboard(
            camera_para_json_dict[camera_id], camera_id)
        camera_para_dict[camera_id] = temp_camera_parameter
    camera_vedo_renderer = VedoRenderer()
    camera_vedo_renderer.set_y_reverse()
    for camera_id in camera_para_dict.keys():
        camera_vedo_renderer.add_camera(camera_para_dict[camera_id])
    if show:
        camera_vedo_renderer.show(with_axis=False, interactive=interactive)


def visualize_dumped_camera_parameter(dumped_dir: str,
                                      interactive: bool = True,
                                      show: bool = True):
    """Visualize all cameras dumped in a directory.

    Args:
        dumped_dir (str):
            Path to the directory.
        interactive (bool, optional):
            Pause and interact with window (True) or
            continue execution (False).
            Defaults to True.
        show (bool, optional):
            Whether to show in a window.
            Defaults to True.
    """
    file_list = os.listdir(dumped_dir)
    camera_para_list = []
    for file_name in file_list:
        file_path = os.path.join(dumped_dir, file_name)
        if not check_path_suffix(file_path, ['.json']):
            continue
        else:
            cam_para = CameraParameter()
            cam_para.load(file_path)
            camera_para_list.append(cam_para)
    camera_vedo_renderer = VedoRenderer()
    camera_vedo_renderer.set_y_reverse()
    for camera_para in camera_para_list:
        camera_vedo_renderer.add_camera(camera_para)
    if show:
        camera_vedo_renderer.show(with_axis=False, interactive=interactive)



import warnings
from typing import List, Optional, Union

import torch
from pytorch3d.io import IO
from pytorch3d.io import load_objs_as_meshes as _load_objs_as_meshes
from pytorch3d.io import save_obj
from pytorch3d.renderer import TexturesUV, TexturesVertex
from pytorch3d.structures import (
    Meshes,
    Pointclouds,
    join_meshes_as_batch,
    join_meshes_as_scene,
    padded_to_list,
)

from .path_utils import prepare_output_path


def join_batch_meshes_as_scene(
    meshes: List[Meshes],
    include_textures: bool = True,
) -> Meshes:
    """Join `meshes` as a scene each batch. Only for Pytorch3D `meshes`. The
    Meshes must share the same batch size, and topology could be different.
    They must all be on the same device. If `include_textures` is true, the
    textures should be the same type, all be None is not accepted. If
    `include_textures` is False, textures are ignored. The return meshes will
    have no textures.

    Args:
        meshes (List[Meshes]): A `list` of `Meshes` with the same batches.
            Required.
        include_textures: (bool) whether to try to join the textures.

    Returns:
        New Meshes which has join different Meshes by each batch.
    """
    for mesh in meshes:
        mesh._verts_list = padded_to_list(mesh.verts_padded(),
                                          mesh.num_verts_per_mesh().tolist())
    num_scene_size = len(meshes)
    num_batch_size = len(meshes[0])
    for i in range(num_scene_size):
        assert len(
            meshes[i]
        ) == num_batch_size, 'Please make sure that the Meshes all have'
        'the same batch size.'
    meshes_all = []
    for j in range(num_batch_size):
        meshes_batch = []
        for i in range(num_scene_size):
            meshes_batch.append(meshes[i][j])
        meshes_all.append(join_meshes_as_scene(meshes_batch, include_textures))
    meshes_final = join_meshes_as_batch(meshes_all, include_textures)
    return meshes_final


def mesh_to_pointcloud_vc(
    meshes: Meshes,
    include_textures: bool = True,
    alpha: float = 1.0,
) -> Pointclouds:
    """Convert PyTorch3D vertex color `Meshes` to `PointClouds`.

    Args:
        meshes (Meshes): input meshes.
        include_textures (bool, optional): Whether include colors.
            Require the texture of input meshes is vertex color.
            Defaults to True.
        alpha (float, optional): transparency.
            Defaults to 1.0.

    Returns:
        Pointclouds: output pointclouds.
    """
    assert isinstance(
        meshes.textures,
        TexturesVertex), 'textures of input meshes should be `TexturesVertex`'
    vertices = meshes.verts_padded()
    if include_textures:
        verts_rgb = meshes.textures.verts_features_padded()
        verts_rgba = torch.cat(
            [verts_rgb,
             torch.ones_like(verts_rgb)[..., 0:1] * alpha], dim=-1)
    else:
        verts_rgba = None
    pointclouds = Pointclouds(points=vertices, features=verts_rgba)
    return pointclouds


def texture_uv2vc(meshes: Meshes) -> Meshes:
    """Convert a Pytorch3D meshes's textures from TexturesUV to TexturesVertex.

    Args:
        meshes (Meshes): input Meshes.

    Returns:
        Meshes: converted Meshes.
    """
    assert isinstance(meshes.textures, TexturesUV)
    device = meshes.device
    vert_uv = meshes.textures.verts_uvs_padded()
    batch_size = vert_uv.shape[0]
    verts_features = []
    num_verts = meshes.verts_padded().shape[1]
    for index in range(batch_size):
        face_uv = vert_uv[index][meshes.textures.faces_uvs_padded()
                                 [index].view(-1)]

        img = meshes.textures._maps_padded[index]
        width, height, _ = img.shape

        face_uv = face_uv * torch.Tensor([width - 1, height - 1
                                          ]).long().to(device)

        face_uv[:, 0] = torch.clip(face_uv[:, 0], 0, width - 1)
        face_uv[:, 1] = torch.clip(face_uv[:, 1], 0, height - 1)
        face_uv = face_uv.long()
        faces = meshes.faces_padded()
        verts_rgb = torch.zeros(1, num_verts, 3).to(device)
        verts_rgb[:, faces.view(-1)] = img[height - 1 - face_uv[:, 1],
                                           face_uv[:, 0]]
        verts_features.append(verts_rgb)
    verts_features = torch.cat(verts_features)

    meshes = meshes.clone()
    meshes.textures = TexturesVertex(verts_features)
    return meshes


def load_objs_as_meshes(files: List[str],
                        device: Optional[Union[torch.device, str]] = None,
                        load_textures: bool = True,
                        **kwargs) -> Meshes:
    if not isinstance(files, list):
        files = [files]
    return _load_objs_as_meshes(
        files=files, device=device, load_textures=load_textures, **kwargs)


def load_plys_as_meshes(
    files: List[str],
    device: Optional[Union[torch.device, str]] = None,
    load_textures: bool = True,
) -> Meshes:
    writer = IO()
    meshes = []
    if not isinstance(files, list):
        files = [files]
    for idx in range(len(files)):
        assert files[idx].endswith('.ply'), 'Please input .ply files.'
        mesh = writer.load_mesh(
            path=files[idx], include_textures=load_textures, device=device)
        meshes.append(mesh)
    meshes = join_meshes_as_batch(meshes, include_textures=load_textures)
    return meshes


def save_meshes_as_plys(files: List[str],
                        meshes: Meshes = None,
                        verts: torch.Tensor = None,
                        faces: torch.Tensor = None,
                        verts_rgb: torch.Tensor = None) -> None:
    """Save meshes as .ply files. Mainly for vertex color meshes.

    Args:
        files (List[str]): Output .ply file list.
        meshes (Meshes, optional): higher priority than
            (verts & faces & verts_rgb). Defaults to None.
        verts (torch.Tensor, optional): lower priority than meshes.
            Defaults to None.
        faces (torch.Tensor, optional): lower priority than meshes.
            Defaults to None.
        verts_rgb (torch.Tensor, optional): lower priority than meshes.
            Defaults to None.
    """
    if meshes is None:
        assert verts is not None and faces is not None, 'Not mesh input.'
        meshes = Meshes(
            verts=verts,
            faces=faces,
            textures=TexturesVertex(
                verts_features=verts_rgb) if verts_rgb is not None else None)
    else:
        if verts is not None or faces is not None or verts_rgb is not None:
            warnings.warn('Redundant input, will use meshes only.')
    assert files is not None
    if not isinstance(files, list):
        files = [files]
    assert len(files) >= len(meshes), 'Not enough output files.'
    writer = IO()
    for idx in range(len(meshes)):
        assert files[idx].endswith('.ply'), 'Please save as .ply files.'
        writer.save_mesh(
            meshes[idx], files[idx], colors_as_uint8=True, binary=False)


def save_meshes_as_objs(files: List[str], meshes: Meshes = None) -> None:
    """Save meshes as .obj files. Pytorch3D will not save vertex color for.

    .obj, please use `save_meshes_as_plys`.

    Args:
        files (List[str]): Output .obj file list.
        meshes (Meshes, optional):
            Defaults to None.
    """
    if not isinstance(files, list):
        files = [files]

    assert len(files) >= len(meshes), 'Not enough output files.'

    for idx in range(len(meshes)):
        prepare_output_path(
            files[idx], allowed_suffix=['.obj'],
            path_type='file'), 'Please save as .obj files.'
        if isinstance(meshes.textures, TexturesUV):
            verts_uvs = meshes.textures.verts_uvs_padded()[idx]
            faces_uvs = meshes.textures.faces_uvs_padded()[idx]
            texture_map = meshes.textures.maps_padded()[idx]
        else:
            verts_uvs = None
            faces_uvs = None
            texture_map = None
        save_obj(
            f=files[idx],
            verts=meshes.verts_padded()[idx],
            faces=meshes.faces_padded()[idx],
            verts_uvs=verts_uvs,
            faces_uvs=faces_uvs,
            texture_map=texture_map)



import warnings
from typing import Iterable, List, Optional, Tuple, Union

import numpy as np
import torch

from mmhuman3d.utils.transforms import ee_to_rotmat, rotmat_to_ee

CAMERA_CONVENTIONS = {
    'pytorch3d': {
        'axis': '-xyz',
        'left_mm_extrinsic': False,
        'view_to_world': False,
        'left_mm_intrinsic': True,
    },
    'pyrender': {
        'axis': 'xy-z',
        'left_mm_extrinsic': True,
        'view_to_world': False,
        'left_mm_intrinsic': True,
    },
    'opengl': {
        'axis': 'xy-z',
        'left_mm_extrinsic': True,
        'view_to_world': False,
        'left_mm_intrinsic': True,
    },
    'open3d': {
        'axis': 'x-yz',
        'left_mm_extrinsic': False,
        'view_to_world': False,
        'left_mm_intrinsic': False,
    },
    'opencv': {
        'axis': 'x-yz',
        'left_mm_extrinsic': True,
        'view_to_world': True,
        'left_mm_intrinsic': True,
    },
    'unity': {
        'axis': 'xyz',
        'left_mm_extrinsic': True,
        'view_to_world': False,
        'left_mm_intrinsic': True,
    },
    'blender': {
        'axis': 'xy-z',
        'left_mm_extrinsic': True,
        'view_to_world': False,
        'left_mm_intrinsic': True,
    },
    'maya': {
        'axis': 'xy-z',
        'left_mm_extrinsic': True,
        'view_to_world': False,
        'left_mm_intrinsic': True,
    }
}


def enc_camera_convention(convention, camera_conventions=CAMERA_CONVENTIONS):
    """convert camera convention to axis direction and order."""
    if convention in camera_conventions:
        convention = camera_conventions[convention]['axis']
    else:
        assert set(convention).issubset(
            {'x', 'y', 'z', '+',
             '-'}), 'Wrong convention string, choose either in'
        f'set({camera_conventions.keys()}) or define by xyz.'
    sign = [1, 1, 1]
    convention = '_' + convention
    count = 0
    axis_order = ''
    for i in range(len(convention)):
        if convention[i] in 'xyz':
            axis_order += convention[i]
            if convention[i - 1] == '-':
                sign[count] *= -1
            count += 1
    return sign, axis_order


def convert_camera_matrix(
    K: Optional[Union[torch.Tensor, np.ndarray]] = None,
    R: Optional[Union[torch.Tensor, np.ndarray]] = None,
    T: Optional[Union[torch.Tensor, np.ndarray]] = None,
    is_perspective: bool = True,
    convention_src: str = 'opencv',
    convention_dst: str = 'pytorch3d',
    in_ndc_src: bool = True,
    in_ndc_dst: bool = True,
    resolution_src: Optional[Union[int, Tuple[int, int], torch.Tensor,
                                   np.ndarray]] = None,
    resolution_dst: Optional[Union[int, Tuple[int, int], torch.Tensor,
                                   np.ndarray]] = None,
    camera_conventions: dict = CAMERA_CONVENTIONS,
) -> Tuple[Union[torch.Tensor, np.ndarray], Union[torch.Tensor, np.ndarray],
           Union[torch.Tensor, np.ndarray]]:
    """Convert the intrinsic matrix K and extrinsic matrix [R|T] from source
    convention to destination convention.

    Args:
        K (Union[torch.Tensor, np.ndarray]): Intrinsic matrix,
            shape should be (batch_size, 4, 4) or (batch_size, 3, 3).
            Will be ignored if None.
        R (Optional[Union[torch.Tensor, np.ndarray]], optional):
            Extrinsic rotation matrix. Shape should be (batch_size, 3, 3).
            Will be identity if None.
            Defaults to None.
        T (Optional[Union[torch.Tensor, np.ndarray]], optional):
            Extrinsic translation matrix. Shape should be (batch_size, 3).
            Will be zeros if None.
            Defaults to None.
        is_perspective (bool, optional): whether is perspective projection.
            Defaults to True.

        _____________________________________________________________________
        # Camera dependent args
        convention_src (str, optional): convention of source camera,
        convention_dst (str, optional): convention of destination camera,

        We define the convention of cameras by the order of right, front and
        up.
        E.g., the first one is pyrender and its convention should be
            '+x+z+y'. '+' could be ignored.
            The second one is opencv and its convention should be '+x-z-y'.
            The third one is pytorch3d and its convention should be '-xzy'.
                    opengl(pyrender)     opencv            pytorch3d
                    y                   z                     y
                    |                  /                      |
                    |                 /                       |
                    |_______x        /________x     x________ |
                    /                |                        /
                   /                 |                       /
                z /                y |                    z /

        in_ndc_src (bool, optional): Whether is the source camera defined
            in ndc.
            Defaults to True.
        in_ndc_dst (bool, optional): Whether is the destination camera defined
            in ndc.
            Defaults to True.

        in camera_convention, we define these args as:
            1). `left_mm_ex` means extrinsic matrix `K` is left matrix
                multiplcation defined.
            2). `left_mm_in` means intrinsic matrix [`R`| `T`] is left
                matrix multiplcation defined.
            3) `view_to_world` means extrinsic matrix [`R`| `T`] is defined
                as view to world.

        resolution_src (Optional[Union[int, Tuple[int, int], torch.Tensor,
            np.ndarray]], optional):
            Source camera image size of (height, width).
            Required if defined in screen.
            Will be square if int.
            Shape should be (2,) if `array` or `tensor`.
            Defaults to None.
        resolution_dst (Optional[Union[int, Tuple[int, int], torch.Tensor,
            np.ndarray]], optional):
            Destination camera image size of (height, width).
            Required if defined in screen.
            Will be square if int.
            Shape should be (2,) if `array` or `tensor`.
            Defaults to None.
        camera_conventions: (dict, optional): `dict` containing
            pre-defined camera convention information.
            Defaults to CAMERA_CONVENTIONS.

    Raises:
        TypeError: K, R, T should all be `torch.Tensor` or `np.ndarray`.

    Returns:
        Tuple[Union[torch.Tensor, None], Union[torch.Tensor, None],
            Union[torch.Tensor, None]]:
            Converted K, R, T matrix of `tensor`.
    """
    convention_dst = convention_dst.lower()
    convention_src = convention_src.lower()

    assert convention_dst in CAMERA_CONVENTIONS
    assert convention_src in CAMERA_CONVENTIONS

    left_mm_ex_src = CAMERA_CONVENTIONS[convention_src].get(
        'left_mm_extrinsic', True)
    view_to_world_src = CAMERA_CONVENTIONS[convention_src].get(
        'view_to_world', False)
    left_mm_in_src = CAMERA_CONVENTIONS[convention_src].get(
        'left_mm_intrinsic', False)

    left_mm_ex_dst = CAMERA_CONVENTIONS[convention_dst].get(
        'left_mm_extrinsic', True)
    view_to_world_dst = CAMERA_CONVENTIONS[convention_dst].get(
        'view_to_world', False)
    left_mm_in_dst = CAMERA_CONVENTIONS[convention_dst].get(
        'left_mm_intrinsic', False)

    sign_src, axis_src = enc_camera_convention(convention_src,
                                               camera_conventions)
    sign_dst, axis_dst = enc_camera_convention(convention_dst,
                                               camera_conventions)
    sign = torch.Tensor(sign_dst) / torch.Tensor(sign_src)

    type_ = []
    for x in [K, R, T]:
        if x is not None:
            type_.append(type(x))
    if len(type_) > 0:
        if not all(x == type_[0] for x in type_):
            raise TypeError('Input type should be the same.')

    use_numpy = False
    if np.ndarray in type_:
        use_numpy = True
    # convert raw matrix to tensor
    if isinstance(K, np.ndarray):
        new_K = torch.Tensor(K)
    elif K is None:
        new_K = None
    elif isinstance(K, torch.Tensor):
        new_K = K.clone()
    else:
        raise TypeError(
            f'K should be `torch.Tensor` or `numpy.ndarray`, type(K): '
            f'{type(K)}')

    if isinstance(R, np.ndarray):
        new_R = torch.Tensor(R).view(-1, 3, 3)
    elif R is None:
        new_R = torch.eye(3, 3)[None]
    elif isinstance(R, torch.Tensor):
        new_R = R.clone().view(-1, 3, 3)
    else:
        raise TypeError(
            f'R should be `torch.Tensor` or `numpy.ndarray`, type(R): '
            f'{type(R)}')

    if isinstance(T, np.ndarray):
        new_T = torch.Tensor(T).view(-1, 3)
    elif T is None:
        new_T = torch.zeros(1, 3)
    elif isinstance(T, torch.Tensor):
        new_T = T.clone().view(-1, 3)
    else:
        raise TypeError(
            f'T should be `torch.Tensor` or `numpy.ndarray`, type(T): '
            f'{type(T)}')

    if axis_dst != axis_src:
        new_R = ee_to_rotmat(
            rotmat_to_ee(new_R, convention=axis_src), convention=axis_dst)

    # convert extrinsic to world_to_view
    if view_to_world_src is True:
        new_R, new_T = convert_world_view(new_R, new_T)

    # right mm to left mm
    if (not left_mm_ex_src) and left_mm_ex_dst:
        new_R *= sign.to(new_R.device)
        new_R = new_R.permute(0, 2, 1)
    # left mm to right mm
    elif left_mm_ex_src and (not left_mm_ex_dst):
        new_R = new_R.permute(0, 2, 1)
        new_R *= sign.to(new_R.device)
    # right_mm to right mm
    elif (not left_mm_ex_dst) and (not left_mm_ex_src):
        new_R *= sign.to(new_R.device)
    # left mm to left mm
    elif left_mm_ex_src and left_mm_ex_dst:
        new_R *= sign.view(3, 1).to(new_R.device)
    new_T *= sign.to(new_T.device)

    # convert extrinsic to as definition
    if view_to_world_dst is True:
        new_R, new_T = convert_world_view(new_R, new_T)

    # in ndc or in screen
    if in_ndc_dst is False and in_ndc_src is True:
        assert resolution_dst is not None, \
            'dst in screen, should specify resolution_dst.'

    if in_ndc_src is False and in_ndc_dst is True:
        assert resolution_src is not None, \
            'src in screen, should specify resolution_dst.'
    if resolution_dst is None:
        resolution_dst = 2.0
    if resolution_src is None:
        resolution_src = 2.0

    if new_K is not None:
        if left_mm_in_src is False and left_mm_in_dst is True:
            new_K = new_K.permute(0, 2, 1)
        if new_K.shape[-2:] == (3, 3):
            new_K = convert_K_3x3_to_4x4(new_K, is_perspective)
        # src in ndc, dst in screen

        if in_ndc_src is True and (in_ndc_dst is False):
            new_K = convert_ndc_to_screen(
                K=new_K,
                is_perspective=is_perspective,
                sign=sign.to(new_K.device),
                resolution=resolution_dst)
        # src in screen, dst in ndc
        elif in_ndc_src is False and in_ndc_dst is True:
            new_K = convert_screen_to_ndc(
                K=new_K,
                is_perspective=is_perspective,
                sign=sign.to(new_K.device),
                resolution=resolution_src)
        # src in ndc, dst in ndc
        elif in_ndc_src is True and in_ndc_dst is True:
            if is_perspective:
                new_K[:, 0, 2] *= sign[0].to(new_K.device)
                new_K[:, 1, 2] *= sign[1].to(new_K.device)
            else:
                new_K[:, 0, 3] *= sign[0].to(new_K.device)
                new_K[:, 1, 3] *= sign[1].to(new_K.device)
        # src in screen, dst in screen
        else:
            pass

        if left_mm_in_src is True and left_mm_in_dst is False:
            new_K = new_K.permute(0, 2, 1)

        num_batch = max(new_K.shape[0], new_R.shape[0], new_T.shape[0])
        if new_K.shape[0] == 1:
            new_K = new_K.repeat(num_batch, 1, 1)
        if new_R.shape[0] == 1:
            new_R = new_R.repeat(num_batch, 1, 1)
        if new_T.shape[0] == 1:
            new_T = new_T.repeat(num_batch, 1)

    if use_numpy:
        if isinstance(new_K, torch.Tensor):
            new_K = new_K.cpu().numpy()
        if isinstance(new_R, torch.Tensor):
            new_R = new_R.cpu().numpy()
        if isinstance(new_T, torch.Tensor):
            new_T = new_T.cpu().numpy()
    return new_K, new_R, new_T


def convert_K_3x3_to_4x4(
        K: Union[torch.Tensor, np.ndarray],
        is_perspective: bool = True) -> Union[torch.Tensor, np.ndarray]:
    """Convert opencv 3x3 intrinsic matrix to 4x4.

    Args:
        K (Union[torch.Tensor, np.ndarray]):
            Input 3x3 intrinsic matrix, left mm defined.
            [[fx,   0,   px],
             [0,   fy,   py],
             [0,    0,   1]]
        is_perspective (bool, optional): whether is perspective projection.
            Defaults to True.

    Raises:
        TypeError: K is not `Tensor` or `array`.
        ValueError: Shape is not (batch, 3, 3) or (3, 3)

    Returns:
        Union[torch.Tensor, np.ndarray]:
            Output intrinsic matrix.
            for perspective:
                [[fx,   0,    px,   0],
                [0,   fy,    py,   0],
                [0,    0,    0,    1],
                [0,    0,    1,    0]]

            for orthographics:
                [[fx,   0,    0,   px],
                [0,   fy,    0,   py],
                [0,    0,    1,    0],
                [0,    0,    0,    1]]
    """
    if isinstance(K, torch.Tensor):
        K = K.clone()
    elif isinstance(K, np.ndarray):
        K = K.copy()

    else:
        raise TypeError('K should be `torch.Tensor` or `numpy.ndarray`, '
                        f'type(K): {type(K)}.')
    if K.shape[-2:] == (4, 4):
        warnings.warn(
            f'shape of K already is {K.shape}, will pass converting.')
        return K
    use_numpy = False
    if K.ndim == 2:
        K = K[None].reshape(-1, 3, 3)
    elif K.ndim == 3:
        K = K.reshape(-1, 3, 3)
    else:
        raise ValueError(f'Wrong ndim of K: {K.ndim}')

    if isinstance(K, np.ndarray):
        use_numpy = True
    if is_perspective:
        if use_numpy:
            K_ = np.zeros((K.shape[0], 4, 4))
        else:
            K_ = torch.zeros(K.shape[0], 4, 4)
        K_[:, :2, :3] = K[:, :2, :3]
        K_[:, 3, 2] = 1
        K_[:, 2, 3] = 1
    else:
        if use_numpy:
            K_ = np.eye(4, 4)[None].repeat(K.shape[0], 0)
        else:
            K_ = torch.eye(4, 4)[None].repeat(K.shape[0], 1, 1)
        K_[:, :2, :2] = K[:, :2, :2]
        K_[:, :2, 3:] = K[:, :2, 2:]
    return K_


def convert_K_4x4_to_3x3(
        K: Union[torch.Tensor, np.ndarray],
        is_perspective: bool = True) -> Union[torch.Tensor, np.ndarray]:
    """Convert opencv 4x4 intrinsic matrix to 3x3.

    Args:
        K (Union[torch.Tensor, np.ndarray]):
            Input 4x4 intrinsic matrix, left mm defined.
            for perspective:
                [[fx,   0,    px,   0],
                [0,   fy,    py,   0],
                [0,    0,    0,    1],
                [0,    0,    1,    0]]

            for orthographics:
                [[fx,   0,    0,   px],
                [0,   fy,    0,   py],
                [0,    0,    1,    0],
                [0,    0,    0,    1]]
        is_perspective (bool, optional): whether is perspective projection.
            Defaults to True.

    Raises:
        TypeError: type K should be `Tensor` or `array`.
        ValueError: Shape is not (batch, 3, 3) or (3, 3).

    Returns:
        Union[torch.Tensor, np.ndarray]:
            Output 3x3 intrinsic matrix, left mm defined.
            [[fx,   0,   px],
             [0,   fy,   py],
             [0,    0,   1]]
    """

    if isinstance(K, torch.Tensor):
        K = K.clone()
    elif isinstance(K, np.ndarray):
        K = K.copy()
    else:
        raise TypeError('K should be `torch.Tensor` or `numpy.ndarray`, '
                        f'type(K): {type(K)}.')
    if K.shape[-2:] == (3, 3):
        warnings.warn(
            f'shape of K already is {K.shape}, will pass converting.')
        return K
    use_numpy = True if isinstance(K, np.ndarray) else False
    if K.ndim == 2:
        K = K[None].reshape(-1, 4, 4)
    elif K.ndim == 3:
        K = K.reshape(-1, 4, 4)
    else:
        raise ValueError(f'Wrong ndim of K: {K.ndim}')

    if use_numpy:
        K_ = np.eye(3, 3)[None].repeat(K.shape[0], 0)
    else:
        K_ = torch.eye(3, 3)[None].repeat(K.shape[0], 1, 1)
    if is_perspective:
        K_[:, :2, :3] = K[:, :2, :3]
    else:
        K_[:, :2, :2] = K[:, :2, :2]
        K_[:, :2, 2:3] = K[:, :2, 3:4]
    return K_


def convert_ndc_to_screen(
        K: Union[torch.Tensor, np.ndarray],
        resolution: Union[int, Tuple[int, int], List[int], torch.Tensor,
                          np.ndarray],
        sign: Optional[Iterable[int]] = None,
        is_perspective: bool = True) -> Union[torch.Tensor, np.ndarray]:
    """Convert intrinsic matrix from ndc to screen.

    Args:
        K (Union[torch.Tensor, np.ndarray]):
            Input 4x4 intrinsic matrix, left mm defined.
        resolution (Union[int, Tuple[int, int], torch.Tensor, np.ndarray]):
            (height, width) of image.
        sign (Optional[Union[Iterable[int]]], optional): xyz axis sign.
            Defaults to None.
        is_perspective (bool, optional): whether is perspective projection.
            Defaults to True.

    Raises:
        TypeError: K should be Tensor or array.
        ValueError: shape of K should be (batch, 4, 4)

    Returns:
        Union[torch.Tensor, np.ndarray]: output intrinsic matrix.
    """
    sign = [1, 1, 1] if sign is None else sign
    if isinstance(K, torch.Tensor):
        K = K.clone()
    elif isinstance(K, np.ndarray):
        K = K.copy()
    else:
        raise TypeError(
            f'K should be `torch.Tensor` or `np.ndarray`, type(K): {type(K)}')
    if K.ndim == 2:
        K = K[None].reshape(-1, 4, 4)
    elif K.ndim == 3:
        K = K.reshape(-1, 4, 4)
    else:
        raise ValueError(f'Wrong ndim of K: {K.ndim}')

    if isinstance(resolution, (int, float)):
        w_dst = h_dst = resolution
    elif isinstance(resolution, (list, tuple)):
        h_dst, w_dst = resolution
    elif isinstance(resolution, (torch.Tensor, np.ndarray)):
        resolution = resolution.reshape(-1, 2)
        h_dst, w_dst = resolution[:, 0], resolution[:, 1]

    aspect_ratio = w_dst / h_dst
    K[:, 0, 0] *= w_dst / 2
    K[:, 1, 1] *= h_dst / 2
    if aspect_ratio > 1:
        K[:, 0, 0] /= aspect_ratio
    else:
        K[:, 1, 1] *= aspect_ratio
    if is_perspective:
        K[:, 0, 2] *= sign[0]
        K[:, 1, 2] *= sign[1]
        K[:, 0, 2] = (K[:, 0, 2] + 1) * (w_dst / 2)
        K[:, 1, 2] = (K[:, 1, 2] + 1) * (h_dst / 2)
    else:
        K[:, 0, 3] *= sign[0]
        K[:, 1, 3] *= sign[1]
        K[:, 0, 3] = (K[:, 0, 3] + 1) * (w_dst / 2)
        K[:, 1, 3] = (K[:, 1, 3] + 1) * (h_dst / 2)
    return K


def convert_screen_to_ndc(
        K: Union[torch.Tensor, np.ndarray],
        resolution: Union[int, Tuple[int, int], torch.Tensor, np.ndarray],
        sign: Optional[Iterable[int]] = None,
        is_perspective: bool = True) -> Union[torch.Tensor, np.ndarray]:
    """Convert intrinsic matrix from screen to ndc.

    Args:
        K (Union[torch.Tensor, np.ndarray]): input intrinsic matrix.
        resolution (Union[int, Tuple[int, int], torch.Tensor, np.ndarray]):
            (height, width) of image.
        sign (Optional[Union[Iterable[int]]], optional): xyz axis sign.
            Defaults to None.
        is_perspective (bool, optional): whether is perspective projection.
            Defaults to True.

    Raises:
        TypeError: K should be Tensor or array.
        ValueError: shape of K should be (batch, 4, 4)

    Returns:
        Union[torch.Tensor, np.ndarray]: output intrinsic matrix.
    """
    if sign is None:
        sign = [1, 1, 1]

    if isinstance(K, torch.Tensor):
        K = K.clone()
    elif isinstance(K, np.ndarray):
        K = K.copy()
    else:
        raise TypeError(
            f'K should be `torch.Tensor` or `np.ndarray`, type(K): {type(K)}')
    if K.ndim == 2:
        K = K[None].reshape(-1, 4, 4)
    elif K.ndim == 3:
        K = K.reshape(-1, 4, 4)
    else:
        raise ValueError(f'Wrong ndim of K: {K.ndim}')

    if isinstance(resolution, (int, float)):
        w_src = h_src = resolution
    elif isinstance(resolution, (list, tuple)):
        h_src, w_src = resolution
    elif isinstance(resolution, (torch.Tensor, np.ndarray)):
        resolution = resolution.reshape(-1, 2)
        h_src, w_src = resolution[:, 0], resolution[:, 1]

    aspect_ratio = w_src / h_src
    K[:, 0, 0] /= w_src / 2
    K[:, 1, 1] /= h_src / 2
    if aspect_ratio > 1:
        K[:, 0, 0] *= aspect_ratio
    else:
        K[:, 1, 1] /= aspect_ratio
    if is_perspective:
        K[:, 0, 2] = K[:, 0, 2] / (w_src / 2) - 1
        K[:, 1, 2] = K[:, 1, 2] / (h_src / 2) - 1
        K[:, 0, 2] *= sign[0]
        K[:, 1, 2] *= sign[1]
    else:
        K[:, 0, 3] = K[:, 0, 3] / (w_src / 2) - 1
        K[:, 1, 3] = K[:, 1, 3] / (h_src / 2) - 1
        K[:, 0, 3] *= sign[0]
        K[:, 1, 3] *= sign[1]
    return K


def convert_world_view(
    R: Union[torch.Tensor, np.ndarray], T: Union[torch.Tensor, np.ndarray]
) -> Tuple[Union[torch.Tensor, np.ndarray], Union[torch.Tensor, np.ndarray]]:
    """Convert between view_to_world and world_to_view defined extrinsic
    matrix.

    Args:
        R (Union[torch.Tensor, np.ndarray]): extrinsic rotation matrix.
            shape should be (batch, 3, 4)
        T (Union[torch.Tensor, np.ndarray]): extrinsic translation matrix.

    Raises:
        TypeError: R and T should be of the same type.

    Returns:
        Tuple[Union[torch.Tensor, np.ndarray], Union[torch.Tensor,
            np.ndarray]]: output R, T.
    """
    if not (type(R) is type(T)):
        raise TypeError(
            f'R: {type(R)}, T: {type(T)} should have the same type.')
    if isinstance(R, torch.Tensor):
        R = R.clone()
        T = T.clone()
        R = R.permute(0, 2, 1)
        T = -(R @ T.view(-1, 3, 1)).view(-1, 3)
    elif isinstance(R, np.ndarray):
        R = R.copy()
        T = T.copy()
        R = R.transpose(0, 2, 1)
        T = -(R @ T.reshape(-1, 3, 1)).reshape(-1, 3)
    else:
        raise TypeError(f'R: {type(R)}, T: {type(T)} should be torch.Tensor '
                        f'or numpy.ndarray.')
    return R, T



"""This script is modified from [PARE](https://github.com/
mkocabas/PARE/tree/master/pare/models/layers).

Original license please see docs/additional_licenses.md.
"""
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.runner.base_module import BaseModule
from torch.nn.modules.utils import _pair

from mmhuman3d.utils.geometry import rot6d_to_rotmat


class LocallyConnected2d(nn.Module):
    """Locally Connected Layer.

    Args:
        in_channels (int):
            the in channel of the features.
        out_channels (int):
            the out channel of the features.
        output_size (List[int]):
            the output size of the features.
        kernel_size (int):
            the size of the kernel.
        stride (int):
            the stride of the kernel.
    Returns:
        attended_features (torch.Tensor):
            attended feature maps
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 output_size,
                 kernel_size,
                 stride,
                 bias=False):
        super(LocallyConnected2d, self).__init__()
        output_size = _pair(output_size)
        self.weight = nn.Parameter(
            torch.randn(1, out_channels, in_channels, output_size[0],
                        output_size[1], kernel_size**2),
            requires_grad=True,
        )
        if bias:
            self.bias = nn.Parameter(
                torch.randn(1, out_channels, output_size[0], output_size[1]),
                requires_grad=True)
        else:
            self.register_parameter('bias', None)
        self.kernel_size = _pair(kernel_size)
        self.stride = _pair(stride)

    def forward(self, x):
        _, c, h, w = x.size()
        kh, kw = self.kernel_size
        dh, dw = self.stride
        x = x.unfold(2, kh, dh).unfold(3, kw, dw)
        x = x.contiguous().view(*x.size()[:-2], -1)
        # Sum in in_channel and kernel_size dims
        out = (x.unsqueeze(1) * self.weight).sum([2, -1])
        if self.bias is not None:
            out += self.bias
        return out


class KeypointAttention(nn.Module):
    """Keypoint Attention Layer.

    Args:
        use_conv (bool):
            whether to use conv for the attended feature map.
            Default: False
        in_channels (List[int]):
            the in channel of shape_cam features and pose features.
            Default: (256, 64)
        out_channels (List[int]):
            the out channel of shape_cam features and pose features.
            Default: (256, 64)
    Returns:
        attended_features (torch.Tensor):
            attended feature maps
    """

    def __init__(self,
                 use_conv=False,
                 in_channels=(256, 64),
                 out_channels=(256, 64),
                 act='softmax',
                 use_scale=False):
        super(KeypointAttention, self).__init__()
        self.use_conv = use_conv
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.act = act
        self.use_scale = use_scale
        if use_conv:
            self.conv1x1_pose = nn.Conv1d(
                in_channels[0], out_channels[0], kernel_size=1)
            self.conv1x1_shape_cam = nn.Conv1d(
                in_channels[1], out_channels[1], kernel_size=1)

    def forward(self, features, heatmaps):
        batch_size, num_joints, height, width = heatmaps.shape

        if self.use_scale:
            scale = 1.0 / np.sqrt(height * width)
            heatmaps = heatmaps * scale

        if self.act == 'softmax':
            normalized_heatmap = F.softmax(
                heatmaps.reshape(batch_size, num_joints, -1), dim=-1)
        elif self.act == 'sigmoid':
            normalized_heatmap = torch.sigmoid(
                heatmaps.reshape(batch_size, num_joints, -1))
        features = features.reshape(batch_size, -1, height * width)

        attended_features = torch.matmul(normalized_heatmap,
                                         features.transpose(2, 1))
        attended_features = attended_features.transpose(2, 1)

        if self.use_conv:
            if attended_features.shape[1] == self.in_channels[0]:
                attended_features = self.conv1x1_pose(attended_features)
            else:
                attended_features = self.conv1x1_shape_cam(attended_features)

        return attended_features


def interpolate(feat, uv):
    """
    Args:
        feat (torch.Tensor): [B, C, H, W] image features
        uv (torch.Tensor): [B, 2, N] uv coordinates
            in the image plane, range [-1, 1]
    Returns:
        samples[:, :, :, 0] (torch.Tensor):
            [B, C, N] image features at the uv coordinates
    """
    if uv.shape[-1] != 2:
        uv = uv.transpose(1, 2)  # [B, N, 2]
    uv = uv.unsqueeze(2)  # [B, N, 1, 2]
    # NOTE: for newer PyTorch, it seems that training
    # results are degraded due to implementation diff in F.grid_sample
    # for old versions, simply remove the aligned_corners argument.
    if int(torch.__version__.split('.')[1]) < 4:
        samples = torch.nn.functional.grid_sample(feat, uv)  # [B, C, N, 1]
    else:
        samples = torch.nn.functional.grid_sample(
            feat, uv, align_corners=True)  # [B, C, N, 1]
    return samples[:, :, :, 0]  # [B, C, N]


def _softmax(tensor, temperature, dim=-1):
    return F.softmax(tensor * temperature, dim=dim)


def softargmax2d(
    heatmaps,
    temperature=None,
    normalize_keypoints=True,
):
    """Softargmax layer for heatmaps."""
    dtype, device = heatmaps.dtype, heatmaps.device
    if temperature is None:
        temperature = torch.tensor(1.0, dtype=dtype, device=device)
    batch_size, num_channels, height, width = heatmaps.shape
    x = torch.arange(
        0, width, device=device,
        dtype=dtype).reshape(1, 1, 1, width).expand(batch_size, -1, height, -1)
    y = torch.arange(
        0, height, device=device,
        dtype=dtype).reshape(1, 1, height, 1).expand(batch_size, -1, -1, width)
    # Should be Bx2xHxW
    points = torch.cat([x, y], dim=1)
    normalized_heatmap = _softmax(
        heatmaps.reshape(batch_size, num_channels, -1),
        temperature=temperature.reshape(1, -1, 1),
        dim=-1)

    # Should be BxJx2
    keypoints = (
        normalized_heatmap.reshape(batch_size, -1, 1, height * width) *
        points.reshape(batch_size, 1, 2, -1)).sum(dim=-1)

    if normalize_keypoints:
        # Normalize keypoints to [-1, 1]
        keypoints[:, :, 0] = (keypoints[:, :, 0] / (width - 1) * 2 - 1)
        keypoints[:, :, 1] = (keypoints[:, :, 1] / (height - 1) * 2 - 1)

    return keypoints, normalized_heatmap.reshape(batch_size, -1, height, width)


class PareHead(BaseModule):

    def __init__(
        self,
        num_joints=24,
        num_input_features=480,
        softmax_temp=1.0,
        num_deconv_layers=3,
        num_deconv_filters=(256, 256, 256),
        num_deconv_kernels=(4, 4, 4),
        num_camera_params=3,
        num_features_smpl=64,
        final_conv_kernel=1,
        pose_mlp_num_layers=1,
        shape_mlp_num_layers=1,
        pose_mlp_hidden_size=256,
        shape_mlp_hidden_size=256,
        bn_momentum=0.1,
        use_heatmaps='part_segm',
        use_keypoint_attention=False,
        use_postconv_keypoint_attention=False,
        keypoint_attention_act='softmax',  # softmax, sigmoid
        use_scale_keypoint_attention=False,
        backbone='hrnet_w32-conv',  # hrnet, resnet
        smpl_mean_params=None,
        deconv_with_bias=False,
    ):
        """PARE parameters regressor head. This class is modified from.

        [PARE](hhttps://github.com/
        mkocabas/PARE/blob/master/pare/models/head/pare_head.py). Original
        license please see docs/additional_licenses.md.

        Args:
            num_joints (int):
                Number of joints, should be 24 for smpl.
            num_input_features (int):
                Number of input featuremap channels.
            softmax_temp (float):
                Softmax tempreture
            num_deconv_layers (int):
                Number of deconvolution layers.
            num_deconv_filters (List[int]):
                Number of filters for each deconvolution layer,
                len(num_deconv_filters) == num_deconv_layers.
            num_deconv_kernels (List[int]):
                Kernel size  for each deconvolution layer,
                len(num_deconv_kernels) == num_deconv_layers.
            num_camera_params (int):
                Number of predicted camera parameter dimension.
            num_features_smpl (int):
                Number of feature map channels.
            final_conv_kernel (int):
                Kernel size for the final deconvolution feature map channels.
            pose_mlp_num_layers (int):
                Number of mpl layers for pose parameter regression.
            shape_mlp_num_layers (int):
                Number of mpl layers for pose parameter regression.
            pose_mlp_hidden_size (int):
                Hidden size for pose mpl layers.
            shape_mlp_hidden_size (int):
                Hidden size for pose mpl layers.
            bn_momemtum (float):
                Momemtum for batch normalization.
            use_heatmaps (str):
                Types of heat maps to use.
            use_keypoint_attention (bool)
                Whether to use attention based on heat maps.
            keypoint_attention_act (str):
                Types of activation function for attention layers.
            use_scale_keypoint_attention (str):
                Whether to scale the attention
                according to the size of the attention map.
            deconv_with_bias (bool)
                Whether to deconv with bias.
            backbone (str):
                Types of the backbone.
            smpl_mean_params (str):
                File name of the mean SMPL parameters
        """

        super(PareHead, self).__init__()
        self.backbone = backbone
        self.num_joints = num_joints
        self.deconv_with_bias = deconv_with_bias
        self.use_heatmaps = use_heatmaps
        self.pose_mlp_num_layers = pose_mlp_num_layers
        self.shape_mlp_num_layers = shape_mlp_num_layers
        self.pose_mlp_hidden_size = pose_mlp_hidden_size
        self.shape_mlp_hidden_size = shape_mlp_hidden_size
        self.use_keypoint_attention = use_keypoint_attention

        self.num_input_features = num_input_features
        self.bn_momentum = bn_momentum
        if self.use_heatmaps == 'part_segm':

            self.use_keypoint_attention = True

        if backbone.startswith('hrnet'):

            self.keypoint_deconv_layers = self._make_conv_layer(
                num_deconv_layers,
                num_deconv_filters,
                (3, ) * num_deconv_layers,
            )
            self.num_input_features = num_input_features
            self.smpl_deconv_layers = self._make_conv_layer(
                num_deconv_layers,
                num_deconv_filters,
                (3, ) * num_deconv_layers,
            )
        else:
            # part branch that estimates 2d keypoints

            conv_fn = self._make_deconv_layer

            self.keypoint_deconv_layers = conv_fn(
                num_deconv_layers,
                num_deconv_filters,
                num_deconv_kernels,
            )
            # reset inplanes to 2048 -> final resnet layer
            self.num_input_features = num_input_features
            self.smpl_deconv_layers = conv_fn(
                num_deconv_layers,
                num_deconv_filters,
                num_deconv_kernels,
            )

        pose_mlp_inp_dim = num_deconv_filters[-1]
        smpl_final_dim = num_features_smpl
        shape_mlp_inp_dim = num_joints * smpl_final_dim

        self.keypoint_final_layer = nn.Conv2d(
            in_channels=num_deconv_filters[-1],
            out_channels=num_joints +
            1 if self.use_heatmaps in ('part_segm',
                                       'part_segm_pool') else num_joints,
            kernel_size=final_conv_kernel,
            stride=1,
            padding=1 if final_conv_kernel == 3 else 0,
        )

        self.smpl_final_layer = nn.Conv2d(
            in_channels=num_deconv_filters[-1],
            out_channels=smpl_final_dim,
            kernel_size=final_conv_kernel,
            stride=1,
            padding=1 if final_conv_kernel == 3 else 0,
        )

        # temperature for softargmax function
        self.register_buffer('temperature', torch.tensor(softmax_temp))
        mean_params = np.load(smpl_mean_params)
        init_pose = torch.from_numpy(mean_params['pose'][:]).unsqueeze(0)
        init_shape = torch.from_numpy(
            mean_params['shape'][:].astype('float32')).unsqueeze(0)
        init_cam = torch.from_numpy(mean_params['cam']).unsqueeze(0)
        self.register_buffer('init_pose', init_pose)
        self.register_buffer('init_shape', init_shape)
        self.register_buffer('init_cam', init_cam)

        self.pose_mlp_inp_dim = pose_mlp_inp_dim
        self.shape_mlp_inp_dim = shape_mlp_inp_dim

        self.shape_mlp = self._get_shape_mlp(output_size=10)
        self.cam_mlp = self._get_shape_mlp(output_size=num_camera_params)

        self.pose_mlp = self._get_pose_mlp(
            num_joints=num_joints, output_size=6)

        self.keypoint_attention = KeypointAttention(
            use_conv=use_postconv_keypoint_attention,
            in_channels=(self.pose_mlp_inp_dim, smpl_final_dim),
            out_channels=(self.pose_mlp_inp_dim, smpl_final_dim),
            act=keypoint_attention_act,
            use_scale=use_scale_keypoint_attention,
        )

    def _get_shape_mlp(self, output_size):
        """mlp layers for shape regression."""
        if self.shape_mlp_num_layers == 1:
            return nn.Linear(self.shape_mlp_inp_dim, output_size)

        module_list = []
        for i in range(self.shape_mlp_num_layers):
            if i == 0:
                module_list.append(
                    nn.Linear(self.shape_mlp_inp_dim,
                              self.shape_mlp_hidden_size))
            elif i == self.shape_mlp_num_layers - 1:
                module_list.append(
                    nn.Linear(self.shape_mlp_hidden_size, output_size))
            else:
                module_list.append(
                    nn.Linear(self.shape_mlp_hidden_size,
                              self.shape_mlp_hidden_size))
        return nn.Sequential(*module_list)

    def _get_pose_mlp(self, num_joints, output_size):
        """mlp layers for pose regression."""
        if self.pose_mlp_num_layers == 1:

            return LocallyConnected2d(
                in_channels=self.pose_mlp_inp_dim,
                out_channels=output_size,
                output_size=[num_joints, 1],
                kernel_size=1,
                stride=1,
            )

        module_list = []
        for i in range(self.pose_mlp_num_layers):
            if i == 0:
                module_list.append(
                    LocallyConnected2d(
                        in_channels=self.pose_mlp_inp_dim,
                        out_channels=self.pose_mlp_hidden_size,
                        output_size=[num_joints, 1],
                        kernel_size=1,
                        stride=1,
                    ))
            elif i == self.pose_mlp_num_layers - 1:
                module_list.append(
                    LocallyConnected2d(
                        in_channels=self.pose_mlp_hidden_size,
                        out_channels=output_size,
                        output_size=[num_joints, 1],
                        kernel_size=1,
                        stride=1,
                    ))
            else:
                module_list.append(
                    LocallyConnected2d(
                        in_channels=self.pose_mlp_hidden_size,
                        out_channels=self.pose_mlp_hidden_size,
                        output_size=[num_joints, 1],
                        kernel_size=1,
                        stride=1,
                    ))
        return nn.Sequential(*module_list)

    def _get_deconv_cfg(self, deconv_kernel):
        """get deconv padding, output padding according to kernel size."""
        if deconv_kernel == 4:
            padding = 1
            output_padding = 0
        elif deconv_kernel == 3:
            padding = 1
            output_padding = 1
        elif deconv_kernel == 2:
            padding = 0
            output_padding = 0

        return deconv_kernel, padding, output_padding

    def _make_conv_layer(self, num_layers, num_filters, num_kernels):
        """make convolution layers."""
        assert num_layers == len(num_filters), \
            'ERROR: num_conv_layers is different len(num_conv_filters)'
        assert num_layers == len(num_kernels), \
            'ERROR: num_conv_layers is different len(num_conv_filters)'
        layers = []
        for i in range(num_layers):
            kernel, padding, output_padding = \
                self._get_deconv_cfg(num_kernels[i])

            planes = num_filters[i]
            layers.append(
                nn.Conv2d(
                    in_channels=self.num_input_features,
                    out_channels=planes,
                    kernel_size=kernel,
                    stride=1,
                    padding=padding,
                    bias=self.deconv_with_bias))
            layers.append(nn.BatchNorm2d(planes, momentum=self.bn_momentum))
            layers.append(nn.ReLU(inplace=True))
            self.num_input_features = planes

        return nn.Sequential(*layers)

    def _make_deconv_layer(self, num_layers, num_filters, num_kernels):
        """make deconvolution layers."""
        assert num_layers == len(num_filters), \
            'ERROR: num_deconv_layers is different len(num_deconv_filters)'
        assert num_layers == len(num_kernels), \
            'ERROR: num_deconv_layers is different len(num_deconv_filters)'

        layers = []
        for i in range(num_layers):
            kernel, padding, output_padding = \
                self._get_deconv_cfg(num_kernels[i])

            planes = num_filters[i]
            layers.append(
                nn.ConvTranspose2d(
                    in_channels=self.num_input_features,
                    out_channels=planes,
                    kernel_size=kernel,
                    stride=2,
                    padding=padding,
                    output_padding=output_padding,
                    bias=self.deconv_with_bias))
            layers.append(nn.BatchNorm2d(planes, momentum=self.bn_momentum))
            layers.append(nn.ReLU(inplace=True))
            # if self.use_self_attention:
            #     layers.append(SelfAttention(planes))
            self.num_input_features = planes

        return nn.Sequential(*layers)

    def forward(self, features):
        batch_size = features.shape[0]

        init_pose = self.init_pose.expand(batch_size, -1)  # N, Jx6
        init_shape = self.init_shape.expand(batch_size, -1)
        init_cam = self.init_cam.expand(batch_size, -1)

        output = {}

        part_feats = self._get_2d_branch_feats(features)

        part_attention = self._get_part_attention_map(part_feats, output)

        smpl_feats = self._get_3d_smpl_feats(features, part_feats)

        point_local_feat, cam_shape_feats = self._get_local_feats(
            smpl_feats, part_attention, output)

        pred_pose, pred_shape, pred_cam = self._get_final_preds(
            point_local_feat, cam_shape_feats, init_pose, init_shape, init_cam)

        pred_rotmat = rot6d_to_rotmat(pred_pose).reshape(batch_size, 24, 3, 3)

        output.update({
            'pred_pose': pred_rotmat,
            'pred_cam': pred_cam,
            'pred_shape': pred_shape,
        })
        return output

    def _get_local_feats(self, smpl_feats, part_attention, output):
        # 1x1 conv
        """get keypoints and camera features from backbone features."""

        cam_shape_feats = self.smpl_final_layer(smpl_feats)

        if self.use_keypoint_attention:
            point_local_feat = self.keypoint_attention(smpl_feats,
                                                       part_attention)
            cam_shape_feats = self.keypoint_attention(cam_shape_feats,
                                                      part_attention)
        else:
            point_local_feat = interpolate(smpl_feats, output['pred_kp2d'])
            cam_shape_feats = interpolate(cam_shape_feats, output['pred_kp2d'])
        return point_local_feat, cam_shape_feats

    def _get_2d_branch_feats(self, features):
        """get part features from backbone features."""
        part_feats = self.keypoint_deconv_layers(features)

        return part_feats

    def _get_3d_smpl_feats(self, features, part_feats):
        """get smpl feature maps from backbone features."""

        smpl_feats = self.smpl_deconv_layers(features)

        return smpl_feats

    def _get_part_attention_map(self, part_feats, output):
        """get attention map from part feature map."""
        heatmaps = self.keypoint_final_layer(part_feats)

        if self.use_heatmaps == 'part_segm':

            output['pred_segm_mask'] = heatmaps
            # remove the the background channel
            heatmaps = heatmaps[:, 1:, :, :]
        else:
            pred_kp2d, _ = softargmax2d(heatmaps, self.temperature)
            output['pred_kp2d'] = pred_kp2d
            output['pred_heatmaps_2d'] = heatmaps
        return heatmaps

    def _get_final_preds(self, pose_feats, cam_shape_feats, init_pose,
                         init_shape, init_cam):
        """get final preds."""
        return self._pare_get_final_preds(pose_feats, cam_shape_feats,
                                          init_pose, init_shape, init_cam)

    def _pare_get_final_preds(self, pose_feats, cam_shape_feats, init_pose,
                              init_shape, init_cam):
        """get final preds."""
        pose_feats = pose_feats.unsqueeze(-1)  #

        if init_pose.shape[-1] == 6:
            # This means init_pose comes from a previous iteration
            init_pose = init_pose.transpose(2, 1).unsqueeze(-1)
        else:
            # This means init pose comes from mean pose
            init_pose = init_pose.reshape(init_pose.shape[0], 6,
                                          -1).unsqueeze(-1)

        shape_feats = cam_shape_feats

        shape_feats = torch.flatten(shape_feats, start_dim=1)

        pred_pose = self.pose_mlp(pose_feats)
        pred_cam = self.cam_mlp(shape_feats)
        pred_shape = self.shape_mlp(shape_feats)

        pred_pose = pred_pose.squeeze(-1).transpose(2, 1)  # N, J, 6
        return pred_pose, pred_shape, pred_cam



import numpy as np
import torch
from torch.nn import functional as F


def batch_rodrigues(theta):
    """Convert axis-angle representation to rotation matrix.

    Args:
        theta: size = [B, 3]
    Returns:
        Rotation matrix corresponding to the quaternion -- size = [B, 3, 3]
    """
    l1norm = torch.norm(theta + 1e-8, p=2, dim=1)
    angle = torch.unsqueeze(l1norm, -1)
    normalized = torch.div(theta, angle)
    angle = angle * 0.5
    v_cos = torch.cos(angle)
    v_sin = torch.sin(angle)
    quat = torch.cat([v_cos, v_sin * normalized], dim=1)
    return quat_to_rotmat(quat)


def quat_to_rotmat(quat):
    """Convert quaternion coefficients to rotation matrix.

    Args:
        quat: size = [B, 4] 4 <===>(w, x, y, z)
    Returns:
        Rotation matrix corresponding to the quaternion -- size = [B, 3, 3]
    """
    norm_quat = quat
    norm_quat = norm_quat / norm_quat.norm(p=2, dim=1, keepdim=True)
    w = norm_quat[:, 0]
    x = norm_quat[:, 1]
    y = norm_quat[:, 2]
    z = norm_quat[:, 3]
    B = quat.size(0)

    w2, x2, y2, z2 = w.pow(2), x.pow(2), y.pow(2), z.pow(2)
    wx, wy, wz = w * x, w * y, w * z
    xy, xz, yz = x * y, x * z, y * z

    rotMat = torch.stack([
        w2 + x2 - y2 - z2, 2 * xy - 2 * wz, 2 * wy + 2 * xz, 2 * wz + 2 * xy,
        w2 - x2 + y2 - z2, 2 * yz - 2 * wx, 2 * xz - 2 * wy, 2 * wx + 2 * yz,
        w2 - x2 - y2 + z2
    ],
                         dim=1).view(B, 3, 3)
    return rotMat


def rot6d_to_rotmat(x):
    """Convert 6D rotation representation to 3x3 rotation matrix.

    Based on Zhou et al., "On the Continuity of Rotation
    Representations in Neural Networks", CVPR 2019
    Input:
        (B,6) Batch of 6-D rotation representations
    Output:
        (B,3,3) Batch of corresponding rotation matrices
    """
    if isinstance(x, torch.Tensor):
        x = x.reshape(-1, 3, 2)
    elif isinstance(x, np.ndarray):
        x = x.view(-1, 3, 2)
    a1 = x[:, :, 0]
    a2 = x[:, :, 1]
    b1 = F.normalize(a1)
    b2 = F.normalize(a2 - torch.einsum('bi,bi->b', b1, a2).unsqueeze(-1) * b1)
    b3 = torch.cross(b1, b2)
    return torch.stack((b1, b2, b3), dim=-1)


def rotation_matrix_to_angle_axis(rotation_matrix):
    """
    This function is borrowed from https://github.com/kornia/kornia
    Convert 3x4 rotation matrix to Rodrigues vector
    Args:
        rotation_matrix (Tensor): rotation matrix.
    Returns:
        Tensor: Rodrigues vector transformation.
    Shape:
        - Input: :math:`(N, 3, 4)`
        - Output: :math:`(N, 3)`
    Example:
        >>> input = torch.rand(2, 3, 4)  # Nx3x4
        >>> output = tgm.rotation_matrix_to_angle_axis(input)  # Nx3
    """
    if rotation_matrix.shape[1:] == (3, 3):
        rot_mat = rotation_matrix.reshape(-1, 3, 3)
        hom = torch.tensor([0, 0, 1],
                           dtype=torch.float32,
                           device=rotation_matrix.device)
        hom = hom.reshape(1, 3, 1).expand(rot_mat.shape[0], -1, -1)
        rotation_matrix = torch.cat([rot_mat, hom], dim=-1)

    quaternion = rotation_matrix_to_quaternion(rotation_matrix)
    aa = quaternion_to_angle_axis(quaternion)
    aa[torch.isnan(aa)] = 0.0
    return aa


def quaternion_to_angle_axis(quaternion: torch.Tensor) -> torch.Tensor:
    """
    This function is borrowed from https://github.com/kornia/kornia
    Convert quaternion vector to angle axis of rotation.
    Adapted from ceres C++ library: ceres-solver/include/ceres/rotation.h
    Args:
        quaternion (torch.Tensor): tensor with quaternions.
    Return:
        torch.Tensor: tensor with angle axis of rotation.
    Shape:
        - Input: :math:`(*, 4)` where `*` means, any number of dimensions
        - Output: :math:`(*, 3)`
    Example:
        >>> quaternion = torch.rand(2, 4)  # Nx4
        >>> angle_axis = tgm.quaternion_to_angle_axis(quaternion)  # Nx3
    """
    if not torch.is_tensor(quaternion):
        raise TypeError('Input type is not a torch.Tensor. Got {}'.format(
            type(quaternion)))

    if not quaternion.shape[-1] == 4:
        raise ValueError(
            'Input must be a tensor of shape Nx4 or 4. Got {}'.format(
                quaternion.shape))
    # unpack input and compute conversion
    q1: torch.Tensor = quaternion[..., 1]
    q2: torch.Tensor = quaternion[..., 2]
    q3: torch.Tensor = quaternion[..., 3]
    sin_squared_theta: torch.Tensor = q1 * q1 + q2 * q2 + q3 * q3

    sin_theta: torch.Tensor = torch.sqrt(sin_squared_theta)
    cos_theta: torch.Tensor = quaternion[..., 0]
    two_theta: torch.Tensor = 2.0 * torch.where(
        cos_theta < 0.0, torch.atan2(-sin_theta, -cos_theta),
        torch.atan2(sin_theta, cos_theta))

    k_pos: torch.Tensor = two_theta / sin_theta
    k_neg: torch.Tensor = 2.0 * torch.ones_like(sin_theta)
    k: torch.Tensor = torch.where(sin_squared_theta > 0.0, k_pos, k_neg)

    angle_axis: torch.Tensor = torch.zeros_like(quaternion)[..., :3]
    angle_axis[..., 0] += q1 * k
    angle_axis[..., 1] += q2 * k
    angle_axis[..., 2] += q3 * k
    return angle_axis


def rotation_matrix_to_quaternion(rotation_matrix, eps=1e-6):
    """
    This function is borrowed from https://github.com/kornia/kornia
    Convert 3x4 rotation matrix to 4d quaternion vector
    This algorithm is based on algorithm described in
    https://github.com/KieranWynn/pyquaternion/blob/master/pyquaternion/quaternion.py#L201
    Args:
        rotation_matrix (Tensor): the rotation matrix to convert.
    Return:
        Tensor: the rotation in quaternion
    Shape:
        - Input: :math:`(N, 3, 4)`
        - Output: :math:`(N, 4)`
    Example:
        >>> input = torch.rand(4, 3, 4)  # Nx3x4
        >>> output = tgm.rotation_matrix_to_quaternion(input)  # Nx4
    """
    if not torch.is_tensor(rotation_matrix):
        raise TypeError('Input type is not a torch.Tensor. Got {}'.format(
            type(rotation_matrix)))

    if len(rotation_matrix.shape) > 3:
        raise ValueError(
            'Input size must be a three dimensional tensor. Got {}'.format(
                rotation_matrix.shape))
    if not rotation_matrix.shape[-2:] == (3, 4):
        raise ValueError(
            'Input size must be a N x 3 x 4  tensor. Got {}'.format(
                rotation_matrix.shape))

    rmat_t = torch.transpose(rotation_matrix, 1, 2)

    mask_d2 = rmat_t[:, 2, 2] < eps

    mask_d0_d1 = rmat_t[:, 0, 0] > rmat_t[:, 1, 1]
    mask_d0_nd1 = rmat_t[:, 0, 0] < -rmat_t[:, 1, 1]

    t0 = 1 + rmat_t[:, 0, 0] - rmat_t[:, 1, 1] - rmat_t[:, 2, 2]
    q0 = torch.stack([
        rmat_t[:, 1, 2] - rmat_t[:, 2, 1], t0,
        rmat_t[:, 0, 1] + rmat_t[:, 1, 0], rmat_t[:, 2, 0] + rmat_t[:, 0, 2]
    ], -1)
    t0_rep = t0.repeat(4, 1).t()

    t1 = 1 - rmat_t[:, 0, 0] + rmat_t[:, 1, 1] - rmat_t[:, 2, 2]
    q1 = torch.stack([
        rmat_t[:, 2, 0] - rmat_t[:, 0, 2], rmat_t[:, 0, 1] + rmat_t[:, 1, 0],
        t1, rmat_t[:, 1, 2] + rmat_t[:, 2, 1]
    ], -1)
    t1_rep = t1.repeat(4, 1).t()

    t2 = 1 - rmat_t[:, 0, 0] - rmat_t[:, 1, 1] + rmat_t[:, 2, 2]
    q2 = torch.stack([
        rmat_t[:, 0, 1] - rmat_t[:, 1, 0], rmat_t[:, 2, 0] + rmat_t[:, 0, 2],
        rmat_t[:, 1, 2] + rmat_t[:, 2, 1], t2
    ], -1)
    t2_rep = t2.repeat(4, 1).t()

    t3 = 1 + rmat_t[:, 0, 0] + rmat_t[:, 1, 1] + rmat_t[:, 2, 2]
    q3 = torch.stack([
        t3, rmat_t[:, 1, 2] - rmat_t[:, 2, 1],
        rmat_t[:, 2, 0] - rmat_t[:, 0, 2], rmat_t[:, 0, 1] - rmat_t[:, 1, 0]
    ], -1)
    t3_rep = t3.repeat(4, 1).t()

    mask_c0 = mask_d2 * mask_d0_d1
    mask_c1 = mask_d2 * ~mask_d0_d1
    mask_c2 = ~mask_d2 * mask_d0_nd1
    mask_c3 = ~mask_d2 * ~mask_d0_nd1
    mask_c0 = mask_c0.view(-1, 1).type_as(q0)
    mask_c1 = mask_c1.view(-1, 1).type_as(q1)
    mask_c2 = mask_c2.view(-1, 1).type_as(q2)
    mask_c3 = mask_c3.view(-1, 1).type_as(q3)

    q = q0 * mask_c0 + q1 * mask_c1 + q2 * mask_c2 + q3 * mask_c3
    q /= torch.sqrt(t0_rep * mask_c0 + t1_rep * mask_c1 +  # noqa
                    t2_rep * mask_c2 + t3_rep * mask_c3)  # noqa
    q *= 0.5
    return q


def perspective_projection(points, rotation, translation, focal_length,
                           camera_center):
    """This function computes the perspective projection of a set of points.

    Input:
        points (bs, N, 3): 3D points
        rotation (bs, 3, 3): Camera rotation
        translation (bs, 3): Camera translation
        focal_length (bs,) or scalar: Focal length
        camera_center (bs, 2): Camera center
    """
    batch_size = points.shape[0]
    K = torch.zeros([batch_size, 3, 3], device=points.device)
    K[:, 0, 0] = focal_length
    K[:, 1, 1] = focal_length
    K[:, 2, 2] = 1.
    K[:, :-1, -1] = camera_center

    # Transform points
    points = torch.einsum('bij,bkj->bki', rotation, points)
    points = points + translation.unsqueeze(1)

    # Apply perspective distortion
    projected_points = points / points[:, :, -1].unsqueeze(-1)

    # Apply camera intrinsics
    projected_points = torch.einsum('bij,bkj->bki', K, projected_points)

    return projected_points[:, :, :-1]


def estimate_translation_np(S,
                            joints_2d,
                            joints_conf,
                            focal_length=5000,
                            img_size=224):
    """Find camera translation that brings 3D joints S closest to 2D the
    corresponding joints_2d.

    Input:
        S: (25, 3) 3D joint locations
        joints: (25, 3) 2D joint locations and confidence
    Returns:
        (3,) camera translation vector
    """

    num_joints = S.shape[0]
    # focal length
    f = np.array([focal_length, focal_length])
    # optical center
    center = np.array([img_size / 2., img_size / 2.])

    # transformations
    Z = np.reshape(np.tile(S[:, 2], (2, 1)).T, -1)
    XY = np.reshape(S[:, 0:2], -1)
    OO = np.tile(center, num_joints)
    F = np.tile(f, num_joints)
    weight2 = np.reshape(np.tile(np.sqrt(joints_conf), (2, 1)).T, -1)

    # least squares
    Q = np.array([
        F * np.tile(np.array([1, 0]), num_joints),
        F * np.tile(np.array([0, 1]), num_joints),
        OO - np.reshape(joints_2d, -1)
    ]).T
    c = (np.reshape(joints_2d, -1) - OO) * Z - F * XY

    # weighted least squares
    W = np.diagflat(weight2)
    Q = np.dot(W, Q)
    c = np.dot(W, c)

    # square matrix
    A = np.dot(Q.T, Q)
    b = np.dot(Q.T, c)

    # solution
    trans = np.linalg.solve(A, b)

    return trans


def estimate_translation(S, joints_2d, focal_length=5000., img_size=224.):
    """Find camera translation that brings 3D joints S closest to 2D the
    corresponding joints_2d.

    Input:
        S: (B, 49, 3) 3D joint locations
        joints: (B, 49, 3) 2D joint locations and confidence
    Returns:
        (B, 3) camera translation vectors
    """

    device = S.device
    # Use only joints 25:49 (GT joints)
    S = S[:, 25:, :].cpu().numpy()
    joints_2d = joints_2d[:, 25:, :].cpu().numpy()
    joints_conf = joints_2d[:, :, -1]
    joints_2d = joints_2d[:, :, :-1]
    trans = np.zeros((S.shape[0], 3), dtype=np.float32)
    # Find the translation for each example in the batch
    for i in range(S.shape[0]):
        S_i = S[i]
        joints_i = joints_2d[i]
        conf_i = joints_conf[i]
        trans[i] = estimate_translation_np(
            S_i,
            joints_i,
            conf_i,
            focal_length=focal_length,
            img_size=img_size)
    return torch.from_numpy(trans).to(device)


def project_points(points_3d, camera, focal_length, img_res):
    """Perform orthographic projection of 3D points using the camera
    parameters, return projected 2D points in image plane.

    Notes:
        batch size: B
        point number: N
    Args:
        points_3d (Tensor([B, N, 3])): 3D points.
        camera (Tensor([B, 3])): camera parameters with the
            3 channel as (scale, translation_x, translation_y)
    Returns:
        points_2d (Tensor([B, N, 2])): projected 2D points
            in image space.
    """
    batch_size = points_3d.shape[0]
    device = points_3d.device
    cam_t = torch.stack([
        camera[:, 1], camera[:, 2], 2 * focal_length /
        (img_res * camera[:, 0] + 1e-9)
    ],
                        dim=-1)
    camera_center = camera.new_zeros([batch_size, 2])
    rot_t = torch.eye(
        3, device=device,
        dtype=points_3d.dtype).unsqueeze(0).expand(batch_size, -1, -1)
    keypoints_2d = perspective_projection(
        points_3d,
        rotation=rot_t,
        translation=cam_t,
        focal_length=focal_length,
        camera_center=camera_center)
    return keypoints_2d



from typing import Optional

import numpy as np
import torch
from smplx import SMPLX as _SMPLX
from smplx.lbs import vertices2joints

from mmhuman3d.core.conventions.keypoints_mapping import (
    convert_kps,
    get_keypoint_num,
)
from mmhuman3d.core.conventions.segmentation import body_segmentation


class SMPLX(_SMPLX):
    """Extension of the official SMPL-X implementation."""

    body_pose_keys = {'global_orient', 'body_pose'}
    full_pose_keys = {
        'global_orient', 'body_pose', 'left_hand_pose', 'right_hand_pose',
        'jaw_pose', 'leye_pose', 'reye_pose'
    }
    NUM_VERTS = 10475
    NUM_FACES = 20908

    def __init__(self,
                 *args,
                 keypoint_src: str = 'smplx',
                 keypoint_dst: str = 'human_data',
                 keypoint_approximate: bool = False,
                 joints_regressor: str = None,
                 extra_joints_regressor: str = None,
                 **kwargs):
        """
        Args:
            *args: extra arguments for SMPL initialization.
            keypoint_src: source convention of keypoints. This convention
                is used for keypoints obtained from joint regressors.
                Keypoints then undergo conversion into keypoint_dst
                convention.
            keypoint_dst: destination convention of keypoints. This convention
                is used for keypoints in the output.
            keypoint_approximate: whether to use approximate matching in
                convention conversion for keypoints.
            joints_regressor: path to joint regressor. Should be a .npy
                file. If provided, replaces the official J_regressor of SMPL.
            extra_joints_regressor: path to extra joint regressor. Should be
                a .npy file. If provided, extra joints are regressed and
                concatenated after the joints regressed with the official
                J_regressor or joints_regressor.
            **kwargs: extra keyword arguments for SMPL initialization.

        Returns:
            None
        """
        super(SMPLX, self).__init__(*args, **kwargs)
        # joints = [JOINT_MAP[i] for i in JOINT_NAMES]
        self.keypoint_src = keypoint_src
        self.keypoint_dst = keypoint_dst
        self.keypoint_approximate = keypoint_approximate

        # override the default SMPL joint regressor if available
        if joints_regressor is not None:
            joints_regressor = torch.tensor(
                np.load(joints_regressor), dtype=torch.float)
            self.register_buffer('joints_regressor', joints_regressor)

        # allow for extra joints to be regressed if available
        if extra_joints_regressor is not None:
            joints_regressor_extra = torch.tensor(
                np.load(extra_joints_regressor), dtype=torch.float)
            self.register_buffer('joints_regressor_extra',
                                 joints_regressor_extra)

        self.num_verts = self.get_num_verts()
        self.num_joints = get_keypoint_num(convention=self.keypoint_dst)
        self.body_part_segmentation = body_segmentation('smplx')

    def forward(self,
                *args,
                return_verts: bool = True,
                return_full_pose: bool = False,
                **kwargs) -> dict:
        """Forward function.

        Args:
            *args: extra arguments for SMPL
            return_verts: whether to return vertices
            return_full_pose: whether to return full pose parameters
            **kwargs: extra arguments for SMPL

        Returns:
            output: contains output parameters and attributes
        """

        kwargs['get_skin'] = True
        smplx_output = super(SMPLX, self).forward(*args, **kwargs)

        if not hasattr(self, 'joints_regressor'):
            joints = smplx_output.joints
        else:
            joints = vertices2joints(self.joints_regressor,
                                     smplx_output.vertices)

        if hasattr(self, 'joints_regressor_extra'):
            extra_joints = vertices2joints(self.joints_regressor_extra,
                                           smplx_output.vertices)
            joints = torch.cat([joints, extra_joints], dim=1)

        joints, joint_mask = convert_kps(
            joints,
            src=self.keypoint_src,
            dst=self.keypoint_dst,
            approximate=self.keypoint_approximate)
        if isinstance(joint_mask, np.ndarray):
            joint_mask = torch.tensor(
                joint_mask, dtype=torch.uint8, device=joints.device)

        batch_size = joints.shape[0]
        joint_mask = joint_mask.reshape(1, -1).expand(batch_size, -1)

        output = dict(
            global_orient=smplx_output.global_orient,
            body_pose=smplx_output.body_pose,
            joints=joints,
            joint_mask=joint_mask,
            keypoints=torch.cat([joints, joint_mask[:, :, None]], dim=-1),
            betas=smplx_output.betas)

        if return_verts:
            output['vertices'] = smplx_output.vertices
        if return_full_pose:
            output['full_pose'] = smplx_output.full_pose

        return output

    @classmethod
    def tensor2dict(cls,
                    full_pose: torch.Tensor,
                    betas: Optional[torch.Tensor] = None,
                    transl: Optional[torch.Tensor] = None,
                    expression: Optional[torch.Tensor] = None) -> dict:
        """Convert full pose tensor to pose dict.

        Args:
            full_pose (torch.Tensor): shape should be (..., 165) or
                (..., 55, 3). All zeros for T-pose.
            betas (Optional[torch.Tensor], optional): shape should be
                (..., 10). The batch num should be 1 or corresponds with
                full_pose.
                Defaults to None.
            transl (Optional[torch.Tensor], optional): shape should be
                (..., 3). The batch num should be 1 or corresponds with
                full_pose.
                Defaults to None.
            expression (Optional[torch.Tensor], optional): shape should
                be (..., 10). The batch num should be 1 or corresponds with
                full_pose.
                Defaults to None.

        Returns:
            dict: dict of smplx pose containing transl & betas.
        """
        NUM_BODY_JOINTS = cls.NUM_BODY_JOINTS
        NUM_HAND_JOINTS = cls.NUM_HAND_JOINTS
        NUM_FACE_JOINTS = cls.NUM_FACE_JOINTS
        NUM_JOINTS = NUM_BODY_JOINTS + 2 * NUM_HAND_JOINTS + NUM_FACE_JOINTS
        full_pose = full_pose.view(-1, (NUM_JOINTS + 1), 3)
        global_orient = full_pose[:, :1]
        body_pose = full_pose[:, 1:NUM_BODY_JOINTS + 1]
        jaw_pose = full_pose[:, NUM_BODY_JOINTS + 1:NUM_BODY_JOINTS + 2]
        leye_pose = full_pose[:, NUM_BODY_JOINTS + 2:NUM_BODY_JOINTS + 3]
        reye_pose = full_pose[:, NUM_BODY_JOINTS + 3:NUM_BODY_JOINTS + 4]
        left_hand_pose = full_pose[:, NUM_BODY_JOINTS + 4:NUM_BODY_JOINTS + 19]
        right_hand_pose = full_pose[:,
                                    NUM_BODY_JOINTS + 19:NUM_BODY_JOINTS + 34]
        batch_size = body_pose.shape[0]
        if betas is not None:
            # squeeze or unsqueeze betas to 2 dims
            betas = betas.view(-1, betas.shape[-1])
            if betas.shape[0] == 1:
                betas = betas.repeat(batch_size, 1)
        else:
            betas = betas
        transl = transl.view(batch_size, -1) if transl is not None else transl
        expression = expression.view(
            batch_size, -1) if expression is not None else torch.zeros(
                batch_size, 10).to(body_pose.device)
        return {
            'betas':
            betas,
            'global_orient':
            global_orient.view(batch_size, 3),
            'body_pose':
            body_pose.view(batch_size, NUM_BODY_JOINTS * 3),
            'left_hand_pose':
            left_hand_pose.view(batch_size, NUM_HAND_JOINTS * 3),
            'right_hand_pose':
            right_hand_pose.view(batch_size, NUM_HAND_JOINTS * 3),
            'transl':
            transl,
            'expression':
            expression,
            'jaw_pose':
            jaw_pose.view(batch_size, 3),
            'leye_pose':
            leye_pose.view(batch_size, 3),
            'reye_pose':
            reye_pose.view(batch_size, 3),
        }

    @classmethod
    def dict2tensor(cls, smplx_dict: dict) -> torch.Tensor:
        """Convert smplx pose dict to full pose tensor.

        Args:
            smplx_dict (dict): smplx pose dict.

        Returns:
            torch: full pose tensor.
        """
        assert cls.body_pose_keys.issubset(smplx_dict)
        for k in smplx_dict:
            if isinstance(smplx_dict[k], np.ndarray):
                smplx_dict[k] = torch.Tensor(smplx_dict[k])
        NUM_BODY_JOINTS = cls.NUM_BODY_JOINTS
        NUM_HAND_JOINTS = cls.NUM_HAND_JOINTS
        NUM_FACE_JOINTS = cls.NUM_FACE_JOINTS
        NUM_JOINTS = NUM_BODY_JOINTS + 2 * NUM_HAND_JOINTS + NUM_FACE_JOINTS
        global_orient = smplx_dict['global_orient'].reshape(-1, 1, 3)
        body_pose = smplx_dict['body_pose'].reshape(-1, NUM_BODY_JOINTS, 3)
        batch_size = global_orient.shape[0]
        jaw_pose = smplx_dict.get('jaw_pose', torch.zeros((batch_size, 1, 3)))
        leye_pose = smplx_dict.get('leye_pose', torch.zeros(
            (batch_size, 1, 3)))
        reye_pose = smplx_dict.get('reye_pose', torch.zeros(
            (batch_size, 1, 3)))
        left_hand_pose = smplx_dict.get(
            'left_hand_pose', torch.zeros((batch_size, NUM_HAND_JOINTS, 3)))
        right_hand_pose = smplx_dict.get(
            'right_hand_pose', torch.zeros((batch_size, NUM_HAND_JOINTS, 3)))
        full_pose = torch.cat([
            global_orient, body_pose,
            jaw_pose.reshape(-1, 1, 3),
            leye_pose.reshape(-1, 1, 3),
            reye_pose.reshape(-1, 1, 3),
            left_hand_pose.reshape(-1, 15, 3),
            right_hand_pose.reshape(-1, 15, 3)
        ],
                              dim=1).reshape(-1, (NUM_JOINTS + 1) * 3)
        return full_pose



PENN_ACTION_KEYPOINTS = [
    'head',
    'left_shoulder',
    'right_shoulder',
    'left_elbow',
    'right_elbow',
    'left_wrist',
    'right_wrist',
    'left_hip',
    'right_hip',
    'left_knee',
    'right_knee',
    'left_ankle',
    'right_ankle',
]



import torch
import torch.nn as nn
import torch.nn.functional as F

from .utils import weight_reduce_loss


def cross_entropy(pred,
                  label,
                  weight=None,
                  reduction='mean',
                  avg_factor=None,
                  class_weight=None,
                  ignore_index=-100):
    """Calculate the CrossEntropy loss.

    Args:
        pred (torch.Tensor): The prediction with shape (N, C), C is the number
            of classes.
        label (torch.Tensor): The learning label of the prediction.
        weight (torch.Tensor, optional): Sample-wise loss weight.
        reduction (str, optional): The method used to reduce the loss.
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.
        class_weight (list[float], optional): The weight for each class.
        ignore_index (int | None): The label index to be ignored.
            If None, it will be set to default value. Default: -100.

    Returns:
        torch.Tensor: The calculated loss
    """
    # The default value of ignore_index is the same as F.cross_entropy
    ignore_index = -100 if ignore_index is None else ignore_index
    # element-wise losses
    loss = F.cross_entropy(
        pred,
        label,
        weight=class_weight,
        reduction='none',
        ignore_index=ignore_index)

    # apply weights and do the reduction
    if weight is not None:
        weight = weight.float()
    loss = weight_reduce_loss(
        loss, weight=weight, reduction=reduction, avg_factor=avg_factor)

    return loss


def _expand_onehot_labels(labels, label_weights, label_channels, ignore_index):
    """Expand onehot labels to match the size of prediction."""
    bin_labels = labels.new_full((labels.size(0), label_channels), 0)
    valid_mask = (labels >= 0) & (labels != ignore_index)
    inds = torch.nonzero(
        valid_mask & (labels < label_channels), as_tuple=False)

    if inds.numel() > 0:
        bin_labels[inds, labels[inds]] = 1

    valid_mask = valid_mask.view(-1, 1).expand(labels.size(0),
                                               label_channels).float()
    if label_weights is None:
        bin_label_weights = valid_mask
    else:
        bin_label_weights = label_weights.view(-1, 1).repeat(1, label_channels)
        bin_label_weights *= valid_mask

    return bin_labels, bin_label_weights


def binary_cross_entropy(pred,
                         label,
                         weight=None,
                         reduction='mean',
                         avg_factor=None,
                         class_weight=None,
                         ignore_index=-100):
    """Calculate the binary CrossEntropy loss.

    Args:
        pred (torch.Tensor): The prediction with shape (N, 1).
        label (torch.Tensor): The learning label of the prediction.
        weight (torch.Tensor, optional): Sample-wise loss weight.
        reduction (str, optional): The method used to reduce the loss.
            Options are "none", "mean" and "sum".
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.
        class_weight (list[float], optional): The weight for each class.
        ignore_index (int | None): The label index to be ignored.
            If None, it will be set to default value. Default: -100.

    Returns:
        torch.Tensor: The calculated loss.
    """
    # The default value of ignore_index is the same as F.cross_entropy
    ignore_index = -100 if ignore_index is None else ignore_index
    if pred.dim() != label.dim():
        label, weight = _expand_onehot_labels(label, weight, pred.size(-1),
                                              ignore_index)

    # weighted element-wise losses
    if weight is not None:
        weight = weight.float()
    loss = F.binary_cross_entropy_with_logits(
        pred, label.float(), pos_weight=class_weight, reduction='none')
    # do the reduction for the weighted loss
    loss = weight_reduce_loss(
        loss, weight, reduction=reduction, avg_factor=avg_factor)

    return loss


def mask_cross_entropy(pred,
                       target,
                       label,
                       reduction='mean',
                       avg_factor=None,
                       class_weight=None,
                       ignore_index=None):
    """Calculate the CrossEntropy loss for masks.

    Args:
        pred (torch.Tensor): The prediction with shape (N, C, *), C is the
            number of classes. The trailing * indicates arbitrary shape.
        target (torch.Tensor): The learning label of the prediction.
        label (torch.Tensor): ``label`` indicates the class label of the mask
            corresponding object. This will be used to select the mask in the
            of the class which the object belongs to when the mask prediction
            if not class-agnostic.
        reduction (str, optional): The method used to reduce the loss.
            Options are "none", "mean" and "sum".
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.
        class_weight (list[float], optional): The weight for each class.
        ignore_index (None): Placeholder, to be consistent with other loss.
            Default: None.

    Returns:
        torch.Tensor: The calculated loss

    Example:
        >>> N, C = 3, 11
        >>> H, W = 2, 2
        >>> pred = torch.randn(N, C, H, W) * 1000
        >>> target = torch.rand(N, H, W)
        >>> label = torch.randint(0, C, size=(N,))
        >>> reduction = 'mean'
        >>> avg_factor = None
        >>> class_weights = None
        >>> loss = mask_cross_entropy(pred, target, label, reduction,
        >>>                           avg_factor, class_weights)
        >>> assert loss.shape == (1,)
    """
    assert ignore_index is None, 'BCE loss does not support ignore_index'
    # TODO: handle these two reserved arguments
    assert reduction == 'mean' and avg_factor is None
    num_rois = pred.size()[0]
    inds = torch.arange(0, num_rois, dtype=torch.long, device=pred.device)
    pred_slice = pred[inds, label].squeeze(1)
    return F.binary_cross_entropy_with_logits(
        pred_slice, target, weight=class_weight, reduction='mean')[None]


class CrossEntropyLoss(nn.Module):

    def __init__(self,
                 use_sigmoid=False,
                 use_mask=False,
                 reduction='mean',
                 class_weight=None,
                 ignore_index=None,
                 loss_weight=1.0):
        """CrossEntropyLoss.

        Args:
            use_sigmoid (bool, optional): Whether the prediction uses sigmoid
                of softmax. Defaults to False.
            use_mask (bool, optional): Whether to use mask cross entropy loss.
                Defaults to False.
            reduction (str, optional): . Defaults to 'mean'.
                Options are "none", "mean" and "sum".
            class_weight (list[float], optional): Weight of each class.
                Defaults to None.
            ignore_index (int | None): The label index to be ignored.
                Defaults to None.
            loss_weight (float, optional): Weight of the loss. Defaults to 1.0.
        """
        super(CrossEntropyLoss, self).__init__()
        assert (use_sigmoid is False) or (use_mask is False)
        self.use_sigmoid = use_sigmoid
        self.use_mask = use_mask
        self.reduction = reduction
        self.loss_weight = loss_weight
        self.class_weight = class_weight
        self.ignore_index = ignore_index

        if self.use_sigmoid:
            self.cls_criterion = binary_cross_entropy
        elif self.use_mask:
            self.cls_criterion = mask_cross_entropy
        else:
            self.cls_criterion = cross_entropy

    def forward(self,
                cls_score,
                label,
                weight=None,
                avg_factor=None,
                reduction_override=None,
                ignore_index=None,
                **kwargs):
        """Forward function.

        Args:
            cls_score (torch.Tensor): The prediction.
            label (torch.Tensor): The learning label of the prediction.
            weight (torch.Tensor, optional): Sample-wise loss weight.
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optional): The method used to reduce the
                loss. Options are "none", "mean" and "sum".
            ignore_index (int | None): The label index to be ignored.
                If not None, it will override the default value. Default: None.
        Returns:
            torch.Tensor: The calculated loss.
        """
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        if ignore_index is None:
            ignore_index = self.ignore_index

        if self.class_weight is not None:
            class_weight = cls_score.new_tensor(
                self.class_weight, device=cls_score.device)
        else:
            class_weight = None
        loss_cls = self.loss_weight * self.cls_criterion(
            cls_score,
            label,
            weight,
            class_weight=class_weight,
            reduction=reduction,
            avg_factor=avg_factor,
            ignore_index=ignore_index,
            **kwargs)
        return loss_cls



import warnings
from typing import Iterable, Optional, Tuple, Union

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from pytorch3d.io.obj_io import load_objs_as_meshes
from pytorch3d.ops import interpolate_face_attributes
from pytorch3d.renderer.mesh import TexturesUV
from pytorch3d.renderer.mesh.rasterizer import (
    MeshRasterizer,
    RasterizationSettings,
)
from pytorch3d.structures import Meshes
from pytorch3d.structures.utils import padded_to_packed

from mmhuman3d.core.cameras.cameras import (
    FoVOrthographicCameras,
    MMCamerasBase,
)
from mmhuman3d.utils.path_utils import check_path_suffix
from .builder import RENDERER
from .utils import array2tensor, rgb2bgr


@RENDERER.register_module(name=['uv_renderer', 'uv', 'UV', 'UVRenderer'])
class UVRenderer(nn.Module):
    """Renderer for SMPL(x) UV map."""

    def __init__(
        self,
        resolution: Tuple[int] = 1024,
        model_type: Optional[str] = 'smpl',
        uv_param_path: Optional[str] = None,
        obj_path: Optional[str] = None,
        device: Union[torch.device, str] = 'cpu',
        threshold_size: int = 512,
        # TODO: Solved the sample bug when the resolution is too small.
        # set threshold_size is just a temporary solution.

        # TODO: add smplx_uv.npz and eval the warping & sampling of smplx
        # model.
    ):
        super().__init__()
        self.threshold_size = threshold_size
        num_verts = {'smpl': 6890, 'smplx': 10475}
        self.NUM_VERTS = num_verts[model_type]
        self.device = device
        self.resolution = (resolution, resolution) if isinstance(
            resolution, int) else resolution
        self.uv_param_path = uv_param_path
        self.obj_path = obj_path
        if uv_param_path is not None:
            check_path_suffix(uv_param_path, allowed_suffix=['npz'])
            param_dict = dict(np.load(uv_param_path))

            verts_uv = torch.Tensor(param_dict['verts_uv'])
            verts_u, verts_v = torch.unbind(verts_uv, -1)
            verts_v_ = 1 - verts_u.unsqueeze(-1)
            verts_u_ = verts_v.unsqueeze(-1)
            self.verts_uv = torch.cat([verts_u_, verts_v_], -1).to(self.device)
            self.faces_uv = torch.LongTensor(param_dict['faces_uv']).to(
                self.device)

            self.NUM_VT = self.verts_uv.shape[0]

            self.faces_tensor = torch.LongTensor(param_dict['faces'].astype(
                np.int64)).to(self.device)
            self.num_faces = self.faces_uv.shape[0]
        elif obj_path is not None:
            check_path_suffix(obj_path, allowed_suffix=['obj'])
            mesh_template = load_objs_as_meshes([obj_path])
            self.faces_uv = mesh_template.textures.faces_uvs_padded()[0].to(
                self.device)
            self.verts_uv = mesh_template.textures.verts_uvs_padded()[0].to(
                self.device)
            self.NUM_VT = self.verts_uv.shape[0]
            self.faces_tensor = mesh_template.faces_padded()[0].to(self.device)
            self.num_faces = self.faces_uv.shape[0]
        self.update_fragments()
        self.update_face_uv_pixel()

        self = self.to(self.device)

    def to(self, device):
        if isinstance(device, str):
            device = torch.device(device)
        self.device = device
        for k in dir(self):
            if isinstance(getattr(self, k), (torch.Tensor)):
                setattr(self, k, getattr(self, k).to(device))
        return self

    def update_fragments(self):
        """Update pix_to_face, bary_coords."""
        rasterizer = MeshRasterizer(
            cameras=FoVOrthographicCameras(
                min_x=1, max_x=0, max_y=1, min_y=0, device=self.device),
            raster_settings=RasterizationSettings(
                blur_radius=0,
                image_size=self.resolution,
                faces_per_pixel=1,
                perspective_correct=False,
            )).to(self.device)
        verts_uv = torch.cat([
            self.verts_uv[None],
            torch.ones(1, self.NUM_VT, 1).to(self.device)
        ], -1)

        fragments = rasterizer(
            Meshes(verts=verts_uv, faces=self.faces_uv[None]))
        self.pix_to_face = fragments.pix_to_face[0, ..., 0]
        self.bary_coords = fragments.bary_coords[0, ..., 0, :]
        self.mask = (self.pix_to_face >= 0).long()

    def update_face_uv_pixel(self):
        """Move the pixels lie on the edges inside the mask, then refine the
        rest points by searching the nearest pixel in the faces it should be
        in."""
        H, W = self.resolution
        device = self.device
        cameras = FoVOrthographicCameras(
            min_x=1, max_x=0, max_y=1, min_y=0, device=self.device)
        verts_uv = torch.cat([
            self.verts_uv[None],
            torch.ones(1, self.NUM_VT, 1).to(self.device)
        ], -1)

        verts_uv_pixel = cameras.transform_points_screen(
            verts_uv, image_size=self.resolution).round().long()[0, ..., :2]
        verts_uv_pixel[..., 0] = torch.clip(
            verts_uv_pixel[..., 0], min=0, max=W - 1)
        verts_uv_pixel[..., 1] = torch.clip(
            verts_uv_pixel[..., 1], min=0, max=H - 1)
        verts_uv_pixel = verts_uv_pixel.long()
        mask = self.mask

        wrong_indexes = torch.where(
            mask[verts_uv_pixel.view(-1, 2)[:, 1],
                 verts_uv_pixel.view(-1, 2)[:, 0]] == 0)[0]
        for wrong_index in wrong_indexes:
            proposed_faces = torch.where(self.faces_uv == wrong_index)[0]
            vert_xy = verts_uv_pixel[wrong_index]
            faces_xy = []
            for face_id in proposed_faces:
                x = torch.where(self.pix_to_face == face_id)[1]
                y = torch.where(self.pix_to_face == face_id)[0]
                if x.shape[0] > 0:
                    face_xy = torch.cat([x.unsqueeze(-1), y.unsqueeze(-1)], -1)
                    faces_xy.append(face_xy)
            if len(faces_xy) > 0:
                faces_xy = torch.cat(faces_xy, 0)
                min_arg = torch.argmin(
                    torch.sqrt(((faces_xy - vert_xy) *
                                (faces_xy - vert_xy)).sum(-1).float()))

                verts_uv_pixel[wrong_index] = faces_xy[min_arg]

        up_bound = ((mask[:-1] - mask[1:]) < 0).long()
        bottom_bound = ((mask[1:] - mask[:-1]) < 0).long()
        left_bound = ((mask[:, :-1] - mask[:, 1:]) < 0).long()
        right_bound = ((mask[:, 1:] - mask[:, :-1]) < 0).long()

        left_bound = torch.cat(
            [left_bound, torch.zeros(H, 1).to(device)], 1).unsqueeze(-1)
        right_bound = torch.cat([torch.zeros(H, 1).to(device), right_bound],
                                1).unsqueeze(-1)
        up_bound = torch.cat([up_bound, torch.zeros(1, W).to(device)],
                             0).unsqueeze(-1)
        bottom_bound = torch.cat([torch.zeros(1, W).to(device), bottom_bound],
                                 0).unsqueeze(-1)

        leftup_corner_ = ((mask[:-1, :-1] - mask[1:, 1:]) < 0).long()
        rightup_corner_ = ((mask[:-1, 1:] - mask[1:, :-1]) < 0).long()
        leftbottom_corner_ = ((mask[1:, :-1] - mask[:-1, 1:]) < 0).long()
        rightbottom_corner_ = ((mask[1:, 1:] - mask[:-1, :-1]) < 0).long()

        leftup_corner = torch.zeros_like(mask).long()
        leftup_corner[:-1, :-1] = leftup_corner_
        leftup_corner = leftup_corner.unsqueeze(-1)

        rightup_corner = torch.zeros_like(mask).long()
        rightup_corner[:-1, 1:] = rightup_corner_
        rightup_corner = rightup_corner.unsqueeze(-1)

        leftbottom_corner = torch.zeros_like(mask).long()
        leftbottom_corner[1:, :-1] = leftbottom_corner_
        leftbottom_corner = leftbottom_corner.unsqueeze(-1)

        rightbottom_corner = torch.zeros_like(mask).long()
        rightbottom_corner[1:, 1:] = rightbottom_corner_
        rightbottom_corner = rightbottom_corner.unsqueeze(-1)

        stride_uv_mask = torch.cat([
            right_bound * -1 + left_bound * 1 + rightbottom_corner * -1 +
            leftbottom_corner * 1 + rightup_corner * -1 + leftup_corner * 1,
            up_bound * 1 + bottom_bound * -1 + rightbottom_corner * -1 +
            leftbottom_corner * -1 + rightup_corner * 1 + leftup_corner * 1
        ], -1).long()

        verts_uv_pixel = verts_uv_pixel + stride_uv_mask[
            verts_uv_pixel.view(-1, 2)[:, 1],
            verts_uv_pixel.view(-1, 2)[:, 0]].view(self.NUM_VT, 2)

        face_uv_pixel = verts_uv_pixel[self.faces_uv]

        face_uv_pixel = face_uv_pixel.long()
        self.face_uv_pixel = face_uv_pixel

    def forward(self,
                verts_attr: Optional[torch.Tensor],
                resolution: Optional[Iterable[int]] = None) -> torch.Tensor:
        """Interpolate the vertex attributes to a map.

        Args:
            verts_attr (Optional[torch.Tensor]): shape should be (N, V, C),
                required.
            resolution (Optional[Iterable[int]], optional): resolution to
                override self.resolution. If None, will use self.resolution.
                Defaults to None.

        Returns:
            torch.Tensor: interpolated maps of (N, H, W, C)
        """
        if verts_attr.ndim == 2:
            verts_attr = verts_attr[None]
        if resolution is not None and resolution != self.resolution:
            self.resolution = resolution
            self.update_fragments()
            self.update_face_uv_pixel()

        bary_coords = self.bary_coords
        pix_to_face = self.pix_to_face

        N, V, C = verts_attr.shape
        assert V == self.NUM_VERTS
        verts_attr = verts_attr.view(N * V, C).to(self.device)
        offset_idx = torch.arange(0, N).long() * (self.NUM_VERTS - 1)
        faces_packed = self.faces_tensor[None].repeat(
            N, 1, 1) + offset_idx.view(-1, 1, 1).to(self.device)
        faces_packed = faces_packed.view(-1, 3)
        face_attr = verts_attr[faces_packed]
        assert face_attr.shape == (N * self.num_faces, 3, C)
        pix_to_face = self.pix_to_face.unsqueeze(0).repeat(N, 1,
                                                           1).unsqueeze(-1)
        bary_coords = self.bary_coords[None].repeat(N, 1, 1, 1).unsqueeze(-2)
        maps_padded = interpolate_face_attributes(
            pix_to_face=pix_to_face.to(self.device),
            barycentric_coords=bary_coords.to(self.device),
            face_attributes=face_attr.to(self.device),
        ).squeeze(-2)
        return maps_padded

    def forward_normal_map(self,
                           meshes: Meshes = None,
                           vertices: torch.Tensor = None,
                           resolution: Optional[Iterable[int]] = None,
                           cameras: MMCamerasBase = None) -> torch.Tensor:
        """Interpolate verts normals to a normal map.

        Args:
            meshes (Meshes): input smpl mesh.
                Will override vertices if both not None.
                Defaults to None.
            vertices (torch.Tensor, optional):
                smpl vertices. Defaults to None.
            resolution (Optional[Iterable[int]], optional): resolution to
                override self.resolution. If None, will use self.resolution.
                Defaults to None.
            cameras (MMCamerasBase, optional):
                cameras to see the mesh.
                Defaults to None.
        Returns:
            torch.Tensor: Normal map of shape (N, H, W, 3)
        """
        if meshes is not None:
            verts_normals = meshes.verts_normals_padded()
        elif meshes is None and vertices is not None:
            meshes = Meshes(
                verts=vertices,
                faces=self.faces_tensor[None].repeat(vertices.shape[0], 1, 1))
            verts_normals = meshes.verts_normals_padded()
        else:
            raise ValueError('No valid input.')
        verts_normals = meshes.verts_normals_padded()
        if cameras:
            verts_normals = cameras.get_world_to_view_transform(
            ).transform_normals(verts_normals)
        normal_map = self.forward(
            verts_attr=verts_normals, resolution=resolution)
        return normal_map

    def forward_uvd_map(self,
                        meshes: Meshes = None,
                        vertices: torch.Tensor = None,
                        resolution: Optional[Iterable[int]] = None,
                        cameras: MMCamerasBase = None) -> torch.Tensor:
        """Interpolate the verts xyz value to a uvd map.

        Args:
            meshes (Meshes): input smpl mesh.
                Defaults to None.
            vertices (torch.Tensor, optional):
                smpl vertices. Will override meshes if both not None.
                Defaults to None.
            resolution (Optional[Iterable[int]], optional): resolution to
                override self.resolution. If None, will use self.resolution.
                Defaults to None.
            cameras (MMCamerasBase, optional):
                cameras to see the mesh.
                Defaults to None.

        Returns:
            torch.Tensor: UVD map of shape (N, H, W, 3)
        """
        if vertices is not None:
            verts_uvd = vertices
        elif vertices is None and meshes is not None:
            verts_uvd = meshes.verts_padded()
        else:
            raise ValueError('No valid input.')
        if cameras:
            verts_uvd = cameras.get_world_to_view_transform(
            ).transform_normals(verts_uvd)
        uvd_map = self.forward(verts_attr=verts_uvd, resolution=resolution)
        return uvd_map

    def vertex_resample(
        self,
        maps_padded: torch.Tensor,
        h_flip: bool = False,
    ) -> torch.Tensor:
        """Resample the vertex attributes from a map.

        Args:
            maps_padded (torch.Tensor): shape should be (N, H, W, C). Required.
            h_flip (bool, optional): whether flip horizontally.
                Defaults to False.

        Returns:
            torch.Tensor: resampled vertex attributes. Shape will be (N, V, C)
        """
        if maps_padded.ndim == 3:
            maps_padded = maps_padded[None]

        if h_flip:
            maps_padded = torch.flip(maps_padded, dims=[2])
        N, H, W, C = maps_padded.shape

        if H < self.threshold_size or W < self.threshold_size:
            maps_padded = F.interpolate(
                maps_padded.permute(0, 3, 1, 2),
                size=(self.threshold_size, self.threshold_size),
                mode='bicubic',
                align_corners=False).permute(0, 2, 3, 1)
            H, W = self.threshold_size, self.threshold_size
        if (H, W) != self.resolution:
            self.resolution = (H, W)
            self.update_fragments()
            self.update_face_uv_pixel()
        offset_idx = torch.arange(0, N).long() * (self.NUM_VERTS - 1)
        faces_packed = self.faces_tensor[None].repeat(
            N, 1, 1) + offset_idx.view(-1, 1, 1).to(self.device)
        faces_packed = faces_packed.view(-1, 3)

        verts_feature_packed = torch.zeros(N * self.NUM_VERTS,
                                           C).to(self.device)

        face_uv_pixel = self.face_uv_pixel.view(-1, 2)
        verts_feature_packed[
            faces_packed] = maps_padded[:, face_uv_pixel[:, 1],
                                        face_uv_pixel[:, 0]].view(
                                            N * self.num_faces, 3, C)
        verts_feature_padded = verts_feature_packed.view(N, self.NUM_VERTS, C)

        return verts_feature_padded

    def wrap_normal(
        self,
        meshes: Meshes,
        normal: torch.Tensor = None,
        normal_map: torch.Tensor = None,
    ) -> Meshes:
        """Warp a normal map or vertex normal to the input meshes.

        Args:
            meshes (Meshes): the input meshes.
            normal (torch.Tensor, optional): vertex normal. Shape should be
                (N, V, 3).
                Defaults to None.
            normal_map (torch.Tensor, optional):
                normal map. Defaults to None.

        Returns:
            Meshes: returned meshes.
        """
        if normal_map is not None and normal is None:
            normal = self.vertex_resample(normal_map)
        elif normal_map is not None and normal is not None:
            normal_map = None
        elif normal_map is None and normal is None:
            warnings.warn('Redundant input, will only take displacement.')
        batch_size = len(meshes)
        if normal.ndim == 2:
            normal = normal[None]
        assert normal.shape[1:] == (self.NUM_VERTS, 3)
        assert normal.shape[0] in [batch_size, 1]

        if normal.shape[0] == 1:
            normal = normal.repeat(batch_size, 1, 1)
        meshes = meshes.clone()

        meshes._set_verts_normals(normal)
        return meshes

    def wrap_displacement(
        self,
        meshes: Meshes,
        displacement: torch.Tensor = None,
        displacement_map: torch.Tensor = None,
    ) -> Meshes:
        """Offset a vertex displacement or displacement_map to the input
        meshes.

        Args:
            meshes (Meshes): the input meshes.
            displacement (torch.Tensor, optional): vertex displacement.
                shape should be (N, V, 3).
                Defaults to None.
            displacement_map (torch.Tensor, optional): displacement_map,
                shape should be (N, H, W, 3).
                Defaults to None.

        Returns:
            Meshes: returned meshes.
        """
        if displacement_map is not None and displacement is None:
            displacement = self.vertex_resample(displacement_map)
        elif displacement_map is not None and displacement is not None:
            displacement_map = None
            warnings.warn('Redundant input, will only take displacement.')
        elif displacement_map is None and displacement is None:
            raise ValueError('No valid input.')
        batch_size = len(meshes)
        if displacement.ndim == 2:
            displacement = displacement[None]
        assert displacement.shape[1] == self.NUM_VERTS
        assert displacement.shape[0] in [batch_size, 1]

        if displacement.shape[0] == 1:
            displacement = displacement.repeat(batch_size, 1, 1)
        C = displacement.shape[-1]
        if C == 1:
            displacement = meshes.verts_normals_padded() * displacement

        displacement = padded_to_packed(displacement)

        meshes = meshes.to(self.device)
        meshes = meshes.offset_verts(displacement)
        return meshes

    def wrap_texture(self,
                     texture_map: torch.Tensor,
                     resolution: Optional[Iterable[int]] = None,
                     mode: Optional[str] = 'bicubic',
                     is_bgr: bool = True) -> Meshes:
        """Wrap a texture map to the input meshes.

        Args:
            texture_map (torch.Tensor): the texture map to be wrapped.
                Shape should be (N, H, W, 3)
            resolution (Optional[Iterable[int]], optional): resolution to
                override self.resolution. If None, will use self.resolution.
                Defaults to None.
            mode (Optional[str], optional): interpolate mode.
                Should be in ['nearest', 'bilinear', 'trilinear', 'bicubic',
                'area'].
                Defaults to 'bicubic'.
            is_bgr (bool, optional): Whether the color channel is BGR.
                Defaults to True.

        Returns:
            Meshes: returned meshes.
        """

        assert texture_map.shape[-1] == 3
        if texture_map.ndim == 3:
            texture_map_padded = texture_map[None]
        elif texture_map.ndim == 4:
            texture_map_padded = texture_map
        else:
            raise ValueError(f'Wrong texture_map shape: {texture_map.shape}.')
        N, H, W, _ = texture_map_padded.shape

        resolution = resolution if resolution is not None else (H, W)

        if resolution != (H, W):
            texture_map_padded = F.interpolate(
                texture_map_padded.view(0, 3, 1, 2), resolution,
                mode=mode).view(0, 2, 3, 1)
        assert texture_map_padded.shape[0] in [N, 1]

        if isinstance(texture_map_padded, np.ndarray):
            texture_map_padded = array2tensor(texture_map_padded)
            is_bgr = True
        if is_bgr:
            texture_map_padded = rgb2bgr(texture_map_padded)

        if texture_map_padded.shape[0] == 1:
            texture_map_padded = texture_map_padded.repeat(N, 1, 1, 1)

        faces_uvs = self.faces_uv[None].repeat(N, 1, 1)
        verts_uvs = self.verts_uv[None].repeat(N, 1, 1)
        textures = TexturesUV(
            faces_uvs=faces_uvs, verts_uvs=verts_uvs, maps=texture_map_padded)
        return textures



import warnings
from typing import Iterable, List, Optional, Tuple, Union

import numpy as np

import mmhuman3d.core.conventions.keypoints_mapping as keypoints_mapping
from mmhuman3d.core.visualization.renderer import Axes3dJointsRenderer
from mmhuman3d.utils import (
    get_different_colors,
    prepare_output_path,
    search_limbs,
)


def _norm_pose(pose_numpy: np.ndarray, min_value: Union[float, int],
               max_value: Union[float, int], mask: Union[np.ndarray, list]):
    """Normalize the poses and make the center close to axis center."""
    assert max_value > min_value
    pose_np_normed = pose_numpy.copy()
    if not mask:
        mask = list(range(pose_numpy.shape[-2]))
    axis_num = 3
    axis_stat = np.zeros(shape=[axis_num, 4])
    for axis_index in range(axis_num):
        axis_data = pose_np_normed[..., mask, axis_index]
        axis_min = np.min(axis_data)
        axis_max = np.max(axis_data)
        axis_mid = (axis_min + axis_max) / 2.0
        axis_span = axis_max - axis_min
        axis_stat[axis_index] = np.asarray(
            (axis_min, axis_max, axis_mid, axis_span))
    target_mid = (max_value + min_value) / 2.0
    max_span = np.max(axis_stat[:, 3])
    target_span = max_value - min_value
    for axis_index in range(axis_num):
        pose_np_normed[..., axis_index] = \
            pose_np_normed[..., axis_index] - \
            axis_stat[axis_index, 2]
    pose_np_normed = pose_np_normed / max_span * target_span
    pose_np_normed = pose_np_normed + target_mid
    return pose_np_normed


def visualize_kp3d(
    kp3d: np.ndarray,
    output_path: Optional[str] = None,
    limbs: Optional[Union[np.ndarray, List[int]]] = None,
    palette: Optional[Iterable[int]] = None,
    data_source: str = 'coco',
    mask: Optional[Union[list, tuple, np.ndarray]] = None,
    start: int = 0,
    end: Optional[int] = None,
    resolution: Union[list, Tuple[int, int]] = (1024, 1024),
    fps: Union[float, int] = 30,
    frame_names: Optional[Union[List[str], str]] = None,
    orbit_speed: Union[float, int] = 0.5,
    value_range: Union[Tuple[int, int], list] = (-100, 100),
    pop_parts: Iterable[str] = (),
    disable_limbs: bool = False,
    return_array: Optional[bool] = None,
    convention: str = 'opencv',
    keypoints_factory: dict = keypoints_mapping.KEYPOINTS_FACTORY,
) -> Union[None, np.ndarray]:
    """Visualize 3d keypoints to a video with matplotlib. Support multi person
    and specified limb connections.

    Args:
        kp3d (np.ndarray): shape could be (f * J * 4/3/2) or
            (f * num_person * J * 4/3/2)
        output_path (str): output video path image folder.
        limbs (Optional[Union[np.ndarray, List[int]]], optional):
            if not specified, the limbs will be searched by search_limbs,
            this option is for free skeletons like BVH file.
            Defaults to None.
        palette (Iterable, optional): specified palette, three int represents
            (B, G, R). Should be tuple or list.
            Defaults to None.
        data_source (str, optional): data source type. Defaults to 'coco'.
            choose in ['coco', 'smplx', 'smpl', 'coco_wholebody',
            'mpi_inf_3dhp', 'mpi_inf_3dhp_test', 'h36m', 'pw3d', 'mpii']
        mask (Optional[Union[list, tuple, np.ndarray]], optional):
            mask to mask out the incorrect points. Defaults to None.
        start (int, optional): start frame index. Defaults to 0.
        end (int, optional): end frame index.
            Could be positive int or negative int or None.
            None represents include all the frames.
            Defaults to None.
        resolution (Union[list, Tuple[int, int]], optional):
            (width, height) of the output video
            will be the same size as the original images if not specified.
            Defaults to None.
        fps (Union[float, int], optional): fps. Defaults to 30.
        frame_names (Optional[Union[List[str], str]], optional): List(should be
            the same as frame numbers) or single string or string format
            (like 'frame%06d')for frame title, no title if None.
            Defaults to None.
        orbit_speed (Union[float, int], optional): orbit speed of camera.
            Defaults to 0.5.
        value_range (Union[Tuple[int, int], list], optional):
            range of axis value. Defaults to (-100, 100).
        pop_parts (Iterable[str], optional): The body part names you do not
            want to visualize. Choose in ['left_eye','right_eye', 'nose',
            'mouth', 'face', 'left_hand', 'right_hand']Defaults to [].
        disable_limbs (bool, optional): whether need to disable drawing limbs.
            Defaults to False.
        return_array (bool, optional): Whether to return images as opencv array
            .If None, an array will be returned when frame number is below 100.
            Defaults to None.
        keypoints_factory (dict, optional): Dict of all the conventions.
            Defaults to KEYPOINTS_FACTORY.
    Raises:
        TypeError: check the type of input keypoints.
        FileNotFoundError: check the output video path.

    Returns:
        Union[None, np.ndarray].
    """
    # check input shape
    if not isinstance(kp3d, np.ndarray):
        raise TypeError(
            f'Input type is {type(kp3d)}, which should be numpy.ndarray.')
    kp3d = kp3d.copy()
    if kp3d.shape[-1] == 2:
        kp3d = np.concatenate([kp3d, np.zeros_like(kp3d)[..., 0:1]], axis=-1)
        warnings.warn(
            'The input array is 2-Dimensional coordinates, will concatenate ' +
            f'zeros to the last axis. The new array shape: {kp3d.shape}')
    elif kp3d.shape[-1] >= 4:
        kp3d = kp3d[..., :3]
        warnings.warn(
            'The input array has more than 3-Dimensional coordinates, will ' +
            'keep only the first 3-Dimensions of the last axis. The new ' +
            f'array shape: {kp3d.shape}')
    if kp3d.ndim == 3:
        kp3d = np.expand_dims(kp3d, 1)
    num_frames = kp3d.shape[0]
    assert kp3d.ndim == 4
    assert kp3d.shape[-1] == 3

    if return_array is None:
        if num_frames > 100:
            return_array = False
        else:
            return_array = True

    # check data_source & mask
    if data_source not in keypoints_factory:
        raise ValueError('Wrong data_source. Should choose in' +
                         f'{list(keypoints_factory.keys())}')
    if mask is not None:
        if not isinstance(mask, np.ndarray):
            mask = np.array(mask).reshape(-1)
        assert mask.shape == (
            len(keypoints_factory[data_source]),
        ), f'mask length should fit with keypoints number \
            {len(keypoints_factory[data_source])}'

    # check the output path
    if output_path is not None:
        prepare_output_path(
            output_path,
            path_type='auto',
            tag='output video',
            allowed_suffix=['.mp4', '.gif', ''])

    # slice the frames
    end = num_frames if end is None else end
    kp3d = kp3d[start:end]
    # norm the coordinates
    if value_range is not None:
        # norm pose location to value_range (70% value range)
        mask_index = np.where(np.array(mask) > 0) if mask is not None else None
        margin_width = abs(value_range[1] - value_range[1]) * 0.15
        pose_np_normed = _norm_pose(kp3d, value_range[0] + margin_width,
                                    value_range[1] - margin_width, mask_index)
        input_pose_np = pose_np_normed
    else:
        input_pose_np = kp3d

    # determine the limb connections and palettes
    if limbs is not None:
        limbs_target, limbs_palette = {
            'body': limbs.tolist() if isinstance(limbs, np.ndarray) else limbs
        }, get_different_colors(len(limbs))
    else:
        limbs_target, limbs_palette = search_limbs(
            data_source=data_source, mask=mask)
    if palette is not None:
        limbs_palette = np.array(palette, dtype=np.uint8)[None]

    # check and pop the pop_parts
    assert set(pop_parts).issubset(
        keypoints_mapping.human_data.HUMAN_DATA_PALETTE.keys(
        )), f'wrong part_names in pop_parts, could only \
        choose in{set(keypoints_mapping.human_data.HUMAN_DATA_PALETTE.keys())}'

    for part_name in pop_parts:
        if part_name in limbs_target:
            limbs_target.pop(part_name)

    # initialize renderer and start render
    renderer = Axes3dJointsRenderer()
    renderer.init_camera(cam_hori_speed=orbit_speed, cam_elev_speed=0.2)
    renderer.set_connections(limbs_target, limbs_palette)
    if isinstance(frame_names, str):
        if '%' in frame_names:
            frame_names = [
                frame_names % index for index in range(input_pose_np.shape[0])
            ]
        else:
            frame_names = [frame_names] * input_pose_np.shape[0]
    image_array = renderer.render_kp3d_to_video(
        input_pose_np,
        output_path,
        convention,
        fps=fps,
        resolution=resolution,
        visual_range=value_range,
        frame_names=frame_names,
        disable_limbs=disable_limbs,
        return_array=return_array)
    return image_array



COCO_WHOLEBODY_KEYPOINTS = [
    'nose',
    'left_eye',
    'right_eye',
    'left_ear',
    'right_ear',
    'left_shoulder',
    'right_shoulder',
    'left_elbow',
    'right_elbow',
    'left_wrist',
    'right_wrist',
    'left_hip',
    'right_hip',
    'left_knee',
    'right_knee',
    'left_ankle',
    'right_ankle',
    'left_bigtoe',
    'left_smalltoe',
    'left_heel',
    'right_bigtoe',
    'right_smalltoe',
    'right_heel',
    'face_contour_1',
    'face_contour_2',
    'face_contour_3',
    'face_contour_4',
    'face_contour_5',
    'face_contour_6',
    'face_contour_7',
    'face_contour_8',
    'face_contour_9',
    'face_contour_10',
    'face_contour_11',
    'face_contour_12',
    'face_contour_13',
    'face_contour_14',
    'face_contour_15',
    'face_contour_16',
    'face_contour_17',
    'right_eyebrow_1',
    'right_eyebrow_2',
    'right_eyebrow_3',
    'right_eyebrow_4',
    'right_eyebrow_5',
    'left_eyebrow_5',
    'left_eyebrow_4',
    'left_eyebrow_3',
    'left_eyebrow_2',
    'left_eyebrow_1',
    'nosebridge_1',
    'nosebridge_2',
    'nosebridge_3',
    'nosebridge_4',
    'nose_1',
    'nose_2',
    'nose_3',
    'nose_4',
    'nose_5',
    'right_eye_1',
    'right_eye_2',
    'right_eye_3',
    'right_eye_4',
    'right_eye_5',
    'right_eye_6',
    'left_eye_4',
    'left_eye_3',
    'left_eye_2',
    'left_eye_1',
    'left_eye_6',
    'left_eye_5',
    'mouth_1',
    'mouth_2',
    'mouth_3',
    'mouth_4',
    'mouth_5',
    'mouth_6',
    'mouth_7',
    'mouth_8',
    'mouth_9',
    'mouth_10',
    'mouth_11',
    'mouth_12',
    'lip_1',
    'lip_2',
    'lip_3',
    'lip_4',
    'lip_5',
    'lip_6',
    'lip_7',
    'lip_8',
    'left_hand_root',
    'left_thumb_1',
    'left_thumb_2',
    'left_thumb_3',
    'left_thumb',
    'left_index_1',
    'left_index_2',
    'left_index_3',
    'left_index',
    'left_middle_1',
    'left_middle_2',
    'left_middle_3',
    'left_middle',
    'left_ring_1',
    'left_ring_2',
    'left_ring_3',
    'left_ring',
    'left_pinky_1',
    'left_pinky_2',
    'left_pinky_3',
    'left_pinky',
    'right_hand_root',
    'right_thumb_1',
    'right_thumb_2',
    'right_thumb_3',
    'right_thumb',
    'right_index_1',
    'right_index_2',
    'right_index_3',
    'right_index',
    'right_middle_1',
    'right_middle_2',
    'right_middle_3',
    'right_middle',
    'right_ring_1',
    'right_ring_2',
    'right_ring_3',
    'right_ring',
    'right_pinky_1',
    'right_pinky_2',
    'right_pinky_3',
    'right_pinky',
]
